@@ -15,7 +15,7 @@ from contextlib import contextmanager
 from collections import OrderedDict
 from datetime import datetime
 from functools import wraps
-from io import BytesIO
+from io import BytesIO, SEEK_END
 from itertools import chain
 
 from ... import __version__, __release_date__
@@ -149,7 +149,7 @@ class Integer(Object, int):
         return int.__repr__(self)
 
     def _bytes(self, document):
-        return int.__str__(self).encode('utf_8')
+        return self._repr().encode('utf_8')
 
 
 class Real(Object, float):
@@ -386,15 +386,14 @@ class Stream(Dictionary):
 
     def direct_bytes(self, document):
         out = bytearray()
-        if self._coder:
-            self._coder.close()
-            self.reset()
+        self.reset()
         if not isinstance(self.filter, PassThrough):
             self['Filter'] = self.filter.name
             if self.filter.params:
                 self['DecodeParms'] = self.filter.params
         if 'Length' in self:
             self['Length'].delete(document)
+        assert self._data.tell() == self._data.seek(0, SEEK_END)
         self['Length'] = Integer(self._data.tell())
         out += super().direct_bytes(document)
         out += b'\nstream\n'
@@ -422,7 +421,9 @@ class Stream(Dictionary):
         return self._data.write(b)
 
     def reset(self):
-        self._coder = None
+        if self._coder:
+            self._coder.close()
+            self._coder = None
 
     def __getattr__(self, name):
         # almost as good as inheriting from BytesIO (which is not possible)
@@ -457,6 +458,8 @@ class Document(dict):
 
     def __init__(self, creator):
         self.catalog = Catalog()
+        self.catalog['PageLabels'] = Dictionary(indirect=True)
+        self.catalog['PageLabels']['Nums'] = Array()
         self.info = Dictionary(indirect=True)
         self.timestamp = time.time()
         self.set_info('Creator', creator)
@@ -466,6 +469,12 @@ class Document(dict):
         self.dests = {}
         self._by_object_id = {}
 
+    def get_page(self, index):
+        for i, page in enumerate(self.catalog['Pages'].pages):
+            if i == index:
+                return page
+        raise IndexError
+
     def register(self, obj):
         try:
             reference = self._by_object_id[id(obj)]
@@ -571,6 +580,10 @@ class Document(dict):
             file.close()
 
 
+class XRefStream(Stream):
+    type = 'XRef'
+
+
 class Catalog(Dictionary):
     type = 'Catalog'
 
@@ -593,6 +606,11 @@ class Pages(Dictionary):
         self['Count'] = Integer(self['Count'] + 1)
         return page
 
+    @property
+    def pages(self):
+        for kid in self['Kids']:
+            yield from kid.object.pages
+
 
 class Page(Dictionary):
     type = 'Page'
@@ -604,6 +622,30 @@ class Page(Dictionary):
         self['MediaBox'] = Array([Integer(0), Integer(0),
                                   Real(width), Real(height)])
 
+    @property
+    def pages(self):
+        yield self
+
+
+DECIMAL_ARABIC = Name('D')
+UPPERCASE_ROMAN = Name('R')
+LOWERCASE_ROMAN = Name('r')
+UPPERCASE_LETTERS = Name('A')
+LOWERCASE_LETTERS = Name('a')
+
+
+class PageLabel(Dictionary):
+    type = 'PageLabel'
+
+    def __init__(self, numbering_style=None, label_prefix=None, start=None):
+        super().__init__(indirect=False)
+        if numbering_style:
+            self['S'] = numbering_style
+        if label_prefix is not None:
+            self['P'] = String(label_prefix)
+        if start is not None:
+            self['St'] = Integer(start)
+
 
 class Outlines(Dictionary):
     type = 'Outlines'
@@ -714,22 +756,25 @@ class Type1Font(Font):
     def register_indirect(self, document, visited=None):
         if id(self) in visited:
             return
-        for key in ('FirstChar', 'LastChar', 'Widths', 'Encoding'):
-            if key in self:
-                self[key].delete(document)
-                del self[key]
-        widths = {}
-        for name, code in chain(self.font.encoding.items(),
-                                self.differences.items()):
-            glyph = self.font._glyphs[name]
-            widths[code] = glyph.width
-        first, last = min(widths), max(widths)
-        self['FirstChar'] = Integer(first)
-        self['LastChar'] = Integer(last)
-        self['Widths'] = Array((Integer(widths[c] if c in widths else 0)
-                                for c in range(first, last + 1)), True)
-        if self.differences:
-            self['Encoding'] = FontEncoding(differences=self.differences)
+        try:
+            self.font
+        except AttributeError:    # this font was parsed from a PDF file
+            pass
+        else:
+            for key in ('FirstChar', 'LastChar', 'Widths', 'Encoding'):
+                if key in self:
+                    self[key].delete(document)
+                    del self[key]
+            widths = {code: self.font._glyphs[name].width
+                      for name, code in chain(self.font.encoding.items(),
+                                              self.differences.items())}
+            first, last = min(widths), max(widths)
+            self['FirstChar'] = Integer(first)
+            self['LastChar'] = Integer(last)
+            self['Widths'] = Array((Integer(widths[c] if c in widths else 0)
+                                    for c in range(first, last + 1)), True)
+            if self.differences:
+                self['Encoding'] = FontEncoding(differences=self.differences)
         return super().register_indirect(document, visited=visited)
 
 
