@@ -20,74 +20,92 @@ Classes representing a document:
 
 """
 
+from pathlib import Path
 
 import datetime
-import os
 import pickle
 import sys
 import time
 
 from collections import OrderedDict
+from copy import copy
 from itertools import count
 
 from . import __version__, __release_date__
+from .attribute import OptionSet
 from .backend import pdf
-from .flowable import RIGHT, LEFT, StaticGroupedFlowables
-from .layout import (Container, ReflowRequired, Chain, PageBreakException,
+from .flowable import StaticGroupedFlowables
+from .language import EN
+from .layout import (Container, ReflowRequired,
                      BACKGROUND, CONTENT, HEADER_FOOTER)
-from .number import NUMBER, format_number
-from .reference import TITLE
-from .structure import NewChapterException, Section
-from .style import DocumentLocationType, Specificity, StyleLog
-from .util import NotImplementedAttribute, RefKeyDictionary
+from .number import format_number
+from .style import StyleLog
+from .util import RefKeyDictionary
+from .warnings import warn
 
 
-__all__ = ['Page', 'DocumentPart', 'DocumentSection', 'Document',
-           'PageOrientation', 'PORTRAIT', 'LANDSCAPE']
+__all__ = ['Page', 'PageOrientation', 'PageType', 'Document', 'DocumentTree']
 
 
 class DocumentTree(StaticGroupedFlowables):
-    def __init__(self, source_file, flowables):
+    """Holds the document's contents as a tree of flowables
+
+    Args:
+        flowables (list[Flowable]): the list of top-level flowables
+        source_file (Path): absolute path of the source file, used to locate
+            images and include in logging and error and warnings.
+        options (Reader): frontend-specific options
+
+    """
+
+    def __init__(self, flowables, source_file=None, options=None):
         super().__init__(flowables)
-        self.source_file = source_file
+        self.source_file = source_file.absolute() if source_file else None
+        self.options = options
 
     @property
     def source_root(self):
-        return os.path.abspath(os.path.dirname(self.source_file))
+        return self.source_file.parent if self.source_file else Path.cwd()
 
 
-class PageOrientation(str):
-    pass
+class PageOrientation(OptionSet):
+    values = 'portrait', 'landscape'
 
 
-PORTRAIT = PageOrientation('portrait')
-LANDSCAPE = PageOrientation('landscape')
+class PageType(OptionSet):
+    values = 'left', 'right', 'any'
 
 
 class Page(Container):
-    """A single page in a document. A :class:`Page` is a :class:`Container`, so
-    other containers can be added as children."""
+    """A single page in a document.
+
+    A :class:`Page` is a :class:`Container`, so other containers can be added
+    as children.
+
+    Args:
+        document_part (DocumentPart): the document part this page is part of
+        number (int): the 1-based index of this page in the document part
+        paper (Paper): determines the dimensions of this page
+        orientation (PageOrientation): the orientation of this page
+
+    """
 
     register_with_parent = False
 
-    def __init__(self, document_part, paper, orientation=PORTRAIT):
-        """Initialize this page as part of `document` (:class:`Document`) with a
-        size defined by `paper` (:class:`Paper`). The page's `orientation` can
-        be either :const:`PORTRAIT` or :const:`LANDSCAPE`."""
+    def __init__(self, document_part, number, paper, orientation='portrait'):
         self._document_part = document_part
+        self.number = number
         self.paper = paper
         self.orientation = orientation
-        if orientation is PORTRAIT:
+        if orientation == PageOrientation.PORTRAIT:
             width, height = paper.width, paper.height
-        elif orientation is LANDSCAPE:
+        elif orientation == PageOrientation.LANDSCAPE:
             width, height = paper.height, paper.width
         document = self.document_part.document
         backend_document = document.backend_document
         self.backend_page = document.backend.Page(backend_document,
-                                                  width, height)
-        self.section = None     # will point to the last section on this page
-        self.overflowed_chains = []
-        self._current_section = {}
+                                                  width, height, self.number,
+                                                  self.number_format)
         self._empty = True
         super().__init__('PAGE', None, 0, 0, width, height)
 
@@ -103,43 +121,30 @@ class Page(Container):
         """Returns the page itself."""
         return self
 
-    def set_current_section(self, section, heading):
-        if heading:
-            if (section.level not in self._current_section
-                    or not self._current_section[section.level][1]):
-                self._current_section[section.level] = section, True
-        elif section.level not in self._current_section:
-            self._current_section[section.level] = section, False
-
     def get_current_section(self, level):
         current_section = None
-        for id, section in ((id, element)
-                            for id, element in self.document.elements.items()
-                            if (isinstance(element, Section)
-                                and element.level == level)):
+        for section in (section for section in self.document._sections
+                        if section.level == level):
+            if section.is_hidden(self):
+                continue
+            section_id = section.get_id(self.document)
             try:
-                first_page = self.document.page_references[id]
+                first_page = self.document.page_elements[section_id]
             except KeyError:
                 break
-            if first_page == self.number:
+            if first_page.document_part is not self.document_part:
+                continue
+            elif first_page.number == self.number:
                 return section
-            elif first_page > self.number:
+            elif first_page.number > self.number:
                 break
-            elif first_page <= self.number:
+            elif first_page.number <= self.number:
                 current_section = section
         return current_section
 
-    @property
-    def document_section(self):
-        return self.document_part.document_section
-
-    @property
-    def number(self):
-        return self.document_section.page_number(self)
-
     @property
     def number_format(self):
-        return self.document_section.page_number_format
+        return self.document_part.page_number_format
 
     @property
     def formatted_number(self):
@@ -175,169 +180,67 @@ class BackendDocumentMetadata(object):
         return instance.backend_document.set_metadata(self.name, value)
 
 
-class DocumentPart(object, metaclass=DocumentLocationType):
-    """Part of a :class:`DocumentSection` that has a specific page template."""
+class PartPageCount(object):
+    def __init__(self):
+        self.count = 0
 
-    header = None
-    footer = None
-    end_at = LEFT
+    def __eq__(self, other):
+        return self.count == other.count
 
-    def __init__(self, document_section,
-                 right_page_template, left_page_template, flowables):
-        self.document_section = document_section
-        self.right_page_template = right_page_template
-        self.left_page_template = left_page_template or right_page_template
-        self.flowable_targets = []
-        self.pages = []
-        if flowables:
-            self.chain = Chain(self)
-            for flowable in flowables:
-                self.chain << flowable
-        else:
-            self.chain = None
+    def __iadd__(self, other):
+        self.count += other
+        return self
 
-    @property
-    def document(self):
-        return self.document_section.document
-
-    @property
-    def number_of_pages(self):
-        return len(self.pages)
-
-    def prepare(self):
-        for flowable_target in self.flowable_targets:
-            flowable_target.prepare(flowable_target)
-
-    def render(self, document_page_count):
-        del self.pages[:]
-        self.add_page(self.first_page())
-        for page in self.pages:
-            try:
-                page.render()
-                break_type = None
-            except NewChapterException as nce:
-                break_type = nce.break_type
-            except PageBreakException as pbe:
-                break_type = None
-            page.place()
-            if self.chain and not self.chain.done:
-                next_page_type = LEFT if page.number % 2 else RIGHT
-                page = self.new_page(next_page_type == break_type)
-                self.add_page(page)     # this grows self.pages!
-        page_count = document_page_count + self.number_of_pages
-        next_page_type = LEFT if page_count % 2 else RIGHT
-        if next_page_type == self.end_at:
-            self.add_page(self.first_page())
-        return self.number_of_pages
-
-    def add_page(self, page):
-        """Append `page` (:class:`Page`) to this :class:`DocumentPart`."""
-        self.pages.append(page)
-
-    def first_page(self):
-        return self.new_page(new_chapter=True)
-
-    def new_page(self, new_chapter, **kwargs):
-        """Called by :meth:`render` with the :class:`Chain`s that need more
-        :class:`Container`s. This method should create a new :class:`Page` which
-        contains a container associated with `chain`."""
-        page_template = (self.left_page_template if len(self.pages) % 2
-                         else self.right_page_template)
-        return page_template.page(self, page_template.name, self.chain,
-                                  new_chapter, **kwargs)
-
-    @classmethod
-    def match(cls, styled, container):
-        if isinstance(container.document_part, cls):
-            return Specificity(0, 1, 0, 0, 0)
-        else:
-            return None
-
-
-class DocumentSection(object, metaclass=DocumentLocationType):
-    """A section of a :class:`Document` that has its own page numbering."""
-
-    parts = NotImplementedAttribute()
-
-    def __init__(self, document, page_number_format=NUMBER):
-        self.document = document
-        self.page_number_format = page_number_format
-        self._parts = [part_class(self) for part_class in self.parts]
-        self.previous_number_of_pages = 0
-
-    @property
-    def number_of_pages(self):
-        return sum(part.number_of_pages for part in self._parts)
-
-    @property
-    def pages(self):
-        return (page for part in self._parts for page in part.pages)
-
-    def page_number(self, this_page):
-        for i, page in enumerate(self.pages, start=1):
-            if this_page is page:
-                return i
-
-    def prepare(self):
-        for part in self._parts:
-            part.prepare()
-
-    def render(self, doc_page_count):
-        section_page_count = 0
-        for part in self._parts:
-            part_page_count = part.render(doc_page_count + section_page_count)
-            section_page_count += part_page_count
-        self.previous_number_of_pages = section_page_count
-        return section_page_count
-
-    @classmethod
-    def match(cls, styled, container):
-        if isinstance(container.document_part.document_section, cls):
-            return Specificity(0, 1, 0, 0, 0)
-        else:
-            return None
 
 
 class Document(object):
-    """A document renders the contents described in an input file onto pages.
-    This is an abstract base class; subclasses should implement :meth:`setup`
-    and :meth:`add_to_chain`."""
+    """Renders a document tree to pages
+
+    Args:
+        document_tree (DocumentTree): a tree of the document's contents
+        stylesheet (StyleSheet): style sheet used to style document elements
+        language (Language): the language to use for standard strings
+        strings (Strings): overrides localized strings provided by `language`
+        backend: the backend used for rendering the document
+
+    """
 
     CREATOR = 'rinohtype v{} ({})'.format(__version__, __release_date__)
 
     CACHE_EXTENSION = '.rtc'
 
-    sections = NotImplementedAttribute()
-
     # FIXME: get backend document metadata from Document metadata
     title = BackendDocumentMetadata('title')
     author = BackendDocumentMetadata('author')
     subject = BackendDocumentMetadata('subject')
     keywords = BackendDocumentMetadata('keywords')
 
-    def __init__(self, document_tree, stylesheet, strings=None,
-                 backend=pdf):
+    def __init__(self, document_tree, stylesheet, language, strings=None,
+                 backend=None):
         """`backend` specifies the backend to use for rendering the document."""
+        super().__init__()
         self._print_version_and_license()
         self.front_matter = []
         self.document_tree = document_tree
         self.stylesheet = stylesheet
+        self.language = language
         self._strings = strings or ()
-        self.backend = backend
+        self.backend = backend or pdf
         self.backend_document = self.backend.Document(self.CREATOR)
         self._flowables = list(id(element)
                                for element in document_tree.elements)
 
-        self.metadata = dict(title='Document Title',
-                             date=datetime.date.today())
+        self.metadata = dict(date=datetime.date.today())
         self.counters = {}             # counters for Headings, Figures, Tables
         self.elements = OrderedDict()  # mapping id's to flowables
         self.ids_by_element = RefKeyDictionary()    # mapping elements to id's
         self.references = {}           # mapping id's to reference data
+        self.page_elements = {}        # mapping id's to pages
         self.page_references = {}      # mapping id's to page numbers
-        self.last_page_references = {}
+        self._sections = []
         self.index_entries = {}
         self._unique_id = 0
+        self.error = False
 
     def _print_version_and_license(self):
         print('rinohtype {} ({})  Copyright (c) Brecht Machiels'
@@ -353,16 +256,22 @@ to the terms of the GNU Affero General Public License version 3.''')
         self._unique_id += 1
         return self._unique_id
 
+    def get_metadata(self, key):
+        return copy(self.metadata.get(key))
+
     def register_element(self, element):
         primary_id = (element.get_id(self, create=False)
                       or self._get_unique_id())
         self.ids_by_element[element] = primary_id
         self.elements[primary_id] = element
+        for id in element.secondary_ids:
+            self.elements[id] = element
         return primary_id
 
     def register_page_reference(self, page, element):
         for id in element.get_ids(self):
-            self.page_references[id] = page.number
+            self.page_elements[id] = page
+            self.page_references[id] = page.formatted_number
 
     def set_reference(self, id, reference_type, value):
         id_references = self.references.setdefault(id, {})
@@ -377,34 +286,45 @@ to the terms of the GNU Affero General Public License version 3.''')
             with open(filename + self.CACHE_EXTENSION, 'rb') as file:
                 prev_number_of_pages, prev_page_references = pickle.load(file)
         except (IOError, TypeError):
-            prev_number_of_pages, prev_page_references = [], {}
+            prev_number_of_pages, prev_page_references = {}, {}
         return prev_number_of_pages, prev_page_references
 
     def _save_cache(self, filename, section_number_of_pages, page_references):
-        """Save the current state of the page references to `<filename>.ptc`"""
-        with open(filename + self.CACHE_EXTENSION, 'wb') as file:
+        """Save the current state of the page references to `<filename>.rtc`"""
+        cache_path = Path(filename).with_suffix(self.CACHE_EXTENSION)
+        with cache_path.open('wb') as file:
             cache = (section_number_of_pages, page_references)
             pickle.dump(cache, file)
 
-    def strings(self, strings_class):
-        for strings in self._strings:
-            if isinstance(strings, strings_class):
-                return strings
-        return strings_class()
+    def get_string(self, strings_class, key):
+        if (strings_class in self._strings
+                and key in self._strings[strings_class]):
+            return self._strings[strings_class][key]
+        try:
+            return self.language.strings[strings_class][key]
+        except KeyError:
+            warn('The {} "{}" string is not defined for {} ({}). Using the '
+                 'English string instead.'.format(strings_class.__name__, key,
+                                                  self.language.name,
+                                                  self.language.code))
+            return EN.strings[strings_class][key]
 
     def render(self, filename_root=None, file=None):
         """Render the document repeatedly until the output no longer changes due
         to cross-references that need some iterations to converge."""
+        self.error = False
+        filename_root = Path(filename_root) if filename_root else None
         if filename_root and file is None:
-            filename = filename_root + self.backend_document.extension
-            file = open(filename, 'wb')
+            extension = self.backend_document.extension
+            filename = filename_root.with_suffix(extension)
+            file = filename.open('wb')
         elif file and filename_root is None:
             filename = getattr(file, 'name', None)
         else:
             raise ValueError("You need to specify either 'filename_root' or "
                              "'file'.")
 
-        def has_converged(section_number_of_pages):
+        def has_converged(part_page_counts):
             """Return `True` if the last rendering iteration converged to a
             stable result.
 
@@ -412,30 +332,30 @@ to the terms of the GNU Affero General Public License version 3.''')
             references to document elements have changed since the previous
             rendering iteration."""
             nonlocal prev_number_of_pages, prev_page_references
-            return (section_number_of_pages == prev_number_of_pages and
+            return (part_page_counts == prev_number_of_pages and
                     self.page_references == prev_page_references)
 
+        fake_container = FakeContainer(self)
         try:
-            self.document_tree.build_document(self)
+            self.document_tree.build_document(fake_container)
             (prev_number_of_pages,
              prev_page_references) = self._load_cache(filename_root)
-            _sections = [section for section in self.sections]
-            for prev_num, section in zip(prev_number_of_pages, _sections):
-                section.previous_number_of_pages = prev_num
+
+            self.part_page_counts = prev_number_of_pages
+            self.prepare(fake_container)
+            self.page_elements.clear()
             self.page_references = prev_page_references.copy()
-            for section in _sections:
-                section.prepare()
-            section_num_pages = self._render_pages(_sections)
-            while not has_converged(section_num_pages):
-                prev_number_of_pages = section_num_pages
+            self.part_page_counts = self._render_pages()
+            while not has_converged(self.part_page_counts):
+                prev_number_of_pages = self.part_page_counts
                 prev_page_references = self.page_references.copy()
                 print('Not yet converged, rendering again...')
                 del self.backend_document
                 self.backend_document = self.backend.Document(self.CREATOR)
-                section_num_pages = self._render_pages(_sections)
+                self.part_page_counts = self._render_pages()
             self.create_outlines()
             if filename:
-                self._save_cache(filename_root, section_num_pages,
+                self._save_cache(filename_root, self.part_page_counts,
                                  self.page_references)
                 self.style_log.write_log(filename_root)
                 print('Writing output: {}'.format(filename))
@@ -443,18 +363,28 @@ to the terms of the GNU Affero General Public License version 3.''')
         finally:
             if filename_root:
                 file.close()
+        return not self.error
 
     def create_outlines(self):
+        """Create an outline in the output file that allows for easy navigation
+        of the document. The outline is a hierarchical tree of all the sections
+        in the document."""
         sections = parent = []
         current_level = 1
         stack = []
-        for section_id, section in ((id, flowable)
-                                    for id, flowable in self.elements.items()
-                                    if isinstance(flowable, Section)):
-            section_number = self.get_reference(section_id, NUMBER)
-            section_title = self.get_reference(section_id, TITLE)
+        fake_container = FakeContainer(self)
+        for section in self._sections:
+            if not section.show_in_toc(fake_container):
+                continue
+            section_id = section.get_id(self, create=False)
+            section_number = self.get_reference(section_id, 'number')
+            section_title = self.get_reference(section_id, 'title')
             if section.level > current_level:
-                assert section.level == current_level + 1
+                if section.level != current_level + 1:
+                    warn("Your document section hierarchy is missing levels. "
+                         "Please report this at https://github.com/brechtm"
+                         "/rinohtype/pull/67")
+                    break
                 stack.append(parent)
                 parent = current
             elif section.level < current_level:
@@ -466,23 +396,34 @@ to the terms of the GNU Affero General Public License version 3.''')
             current_level = section.level
         self.backend_document.create_outlines(sections)
 
-    def _render_pages(self, _sections):
+    def _render_pages(self):
         """Render the complete document once and return the number of pages
         rendered."""
         self.style_log = StyleLog(self.stylesheet)
         self.floats = set()
         self.placed_footnotes = set()
-        section_page_counts = []
         self._start_time = time.time()
-        for section in _sections:
-            section_page_count = section.render(sum(section_page_counts))
-            section_page_counts.append(section_page_count)
-        sys.stdout.write('\n')     # for the progress indicator
-        return section_page_counts
 
+        part_page_counts = {}
+        part_page_count = PartPageCount()
+        last_number_format = None
+        for part_template in self.part_templates:
+            part = part_template.document_part(self)
+            if part is None:
+                continue
+            if part_template.page_number_format != last_number_format:
+                part_page_count = PartPageCount()
+            part_page_count += part.render(part_page_count.count + 1)
+            part_page_counts[part_template.name] = part_page_count
+            last_number_format = part_template.page_number_format
+        sys.stdout.write('\n')     # for the progress indicator
+        return part_page_counts
+
+    PROGRESS_TEMPLATE = \
+        '\r{:3d}% [{}{}] ETA {:02d}:{:02d} ({:02d}:{:02d}) page {}'
     PROGRESS_BAR_WIDTH = 40
 
-    def progress(self, flowable):
+    def progress(self, flowable, container):
         try:
             index = self._flowables.index(id(flowable))
         except ValueError:
@@ -493,9 +434,15 @@ to the terms of the GNU Affero General Public License version 3.''')
             passed = int(time_passed)
             eta = int(time_passed / percent * (100 - percent))
             filled = int(self.PROGRESS_BAR_WIDTH * percent / 100)
-            sys.stdout.write('\r{:3d}% [{}{}] ETA {:02d}:{:02d} ({:02d}:{:02d})'
+            sys.stdout.write(self.PROGRESS_TEMPLATE
                              .format(int(percent), filled * '=',
                                      (self.PROGRESS_BAR_WIDTH - filled) * ' ',
                                      eta // 60, eta % 60,
-                                     passed // 60, passed % 60))
+                                     passed // 60, passed % 60,
+                                     container.page.formatted_number))
             sys.stdout.flush()
+
+
+class FakeContainer(object):    # TODO: clean up
+    def __init__(self, document):
+        self.document = document
