@@ -7,28 +7,24 @@
 
 
 import re
-import unicodedata
 
 from datetime import datetime
 
 import rinoh as rt
 
-from . import (DocutilsInlineNode, DocutilsNode,
-               DocutilsBodyNode, DocutilsBodySubNode,
+from . import (DocutilsInlineNode, DocutilsNode, DocutilsBodyNode,
                DocutilsGroupingNode, DocutilsDummyNode)
 from ...dimension import DimensionUnit, INCH, CM, MM, PT, PICA, PERCENT
 from ...util import intersperse
 
 
-# TODO: support for the following nodes is still missing
+# Support for the following nodes is still missing
 # (http://docutils.sourceforge.net/docs/ref/doctree.html)
-# - abbreviation
-# - acronym
-# - container
-# - decoration / header / footer
+# - abbreviation? (not exposed in default reStructuredText; Sphinx has :abbr:)
+# - acronym (not exposed in default reStructuredText?)
 # - math / math_block
-# - pending
-# - substitution_reference
+# - pending (should not appear in final doctree?)
+# - substitution_reference (should not appear in final doctree?)
 
 
 class Text(DocutilsInlineNode):
@@ -39,7 +35,18 @@ class Text(DocutilsInlineNode):
 
 
 class Inline(DocutilsInlineNode):
-    pass
+    style = None
+    class_styles = {}
+
+    @property
+    def style_from_class(self):
+        for cls in self.get('classes'):
+            if cls in self.class_styles:
+                return self.class_styles[cls]
+        return self.style
+
+    def build_styled_text(self):
+        return self.process_content(style=self.style_from_class)
 
 
 class Document(DocutilsBodyNode):
@@ -52,6 +59,22 @@ class DocInfo(DocutilsBodyNode):
         return rt.SetMetadataFlowable(**doc_info)
 
 
+class Decoration(DocutilsGroupingNode):
+    pass
+
+
+class Header(DocutilsBodyNode):
+    def build_flowable(self):
+        return rt.WarnFlowable('Docutils header nodes are ignored. Please '
+                               'configure your document template instead.')
+
+
+class Footer(DocutilsBodyNode):
+    def build_flowable(self):
+        return rt.WarnFlowable('Docutils footer nodes are ignored. Please '
+                               'configure your document template instead.')
+
+
 # bibliographic elements
 
 class DocInfoField(DocutilsInlineNode):
@@ -181,8 +204,10 @@ class Section(DocutilsGroupingNode):
 
 
 class Paragraph(DocutilsBodyNode):
+    style = None
+
     def build_flowable(self):
-        return rt.Paragraph(super().process_content())
+        return rt.Paragraph(self.process_content(), style=self.style)
 
 
 class Compound(DocutilsGroupingNode):
@@ -271,24 +296,20 @@ class Generated(DocutilsInlineNode):
         return self.process_content()
 
 
-class Emphasis(DocutilsInlineNode):
-    def build_styled_text(self):
-        return rt.SingleStyledText(self.text, style='emphasis')
+class Emphasis(Inline):
+    style = 'emphasis'
 
 
-class Strong(DocutilsInlineNode):
-    def build_styled_text(self):
-        return rt.SingleStyledText(self.text, style='strong')
+class Strong(Inline):
+    style = 'strong'
 
 
 class Title_Reference(DocutilsInlineNode):
-    def build_styled_text(self):
-        return rt.SingleStyledText(self.text, style='title reference')
+    style = 'title reference'
 
 
-class Literal(DocutilsInlineNode):
-    def build_styled_text(self):
-        return rt.SingleStyledText(self.text, style='monospaced')
+class Literal(Inline):
+    style = 'monospaced'
 
 
 class Superscript(DocutilsInlineNode):
@@ -329,19 +350,30 @@ class Block_Quote(DocutilsGroupingNode):
 
 
 class Attribution(Paragraph):
-    def build_flowable(self):
-        return rt.Paragraph('\N{EM DASH}' + self.process_content(),
-                            style='attribution')
+    style = 'attribution'
+
+    def process_content(self, style=None):
+        return '\N{EM DASH}' + super().process_content(style)
 
 
-class Line_Block(DocutilsGroupingNode):
+class Line_Block(Paragraph):
     style = 'line block'
 
+    def _process_block(self, line_block):
+        for child in line_block.getchildren():
+            try:
+                yield child.styled_text()
+            except AttributeError:
+                for line in self._process_block(child):
+                    yield rt.Tab() + line
 
-class Line(DocutilsBodyNode):
-    def build_flowable(self):
-        return rt.Paragraph(self.process_content() or '\n',
-                            style='line block line')
+    def process_content(self, style=None):
+        lines = self._process_block(self)
+        return intersperse(lines, rt.Newline())
+
+
+class Line(DocutilsInlineNode):
+    pass
 
 
 class Doctest_Block(DocutilsBodyNode):
@@ -350,22 +382,28 @@ class Doctest_Block(DocutilsBodyNode):
 
 
 class Reference(DocutilsBodyNode, DocutilsInlineNode):
-    def build_styled_text(self):
+    @property
+    def annotation(self):
         if self.get('refid'):
-            link = rt.NamedDestinationLink(self.get('refid'))
-            style = 'internal link'
+            return rt.NamedDestinationLink(self.get('refid'))
         elif self.get('refuri'):
-            link = rt.HyperLink(self.get('refuri'))
-            style = 'external link'
-        else:
-            return rt.MixedStyledText(self.process_content(),
-                                      style='broken link')
-        return rt.AnnotatedText(self.process_content(), link, style=style)
+            return rt.HyperLink(self.get('refuri'))
+
+    def build_styled_text(self):
+        annotation = self.annotation
+        content = self.process_content()
+        if annotation is None:
+            return rt.MixedStyledText(content, style='broken link')
+        style = ('external link' if annotation.type == 'URI'
+                 else 'internal link')
+        return rt.AnnotatedText(content, annotation, style=style)
 
     def build_flowable(self):
         children = self.getchildren()
         assert len(children) == 1
-        return self.image.flowable()
+        image = self.image.flowable()
+        image.annotation = self.annotation
+        return image
 
 
 class Footnote(DocutilsBodyNode):
@@ -437,21 +475,21 @@ class List_Item(DocutilsGroupingNode):
     pass
 
 
-class Definition_List(DocutilsBodyNode):
+class Definition_List(DocutilsGroupingNode):
+    grouped_flowables_class = rt.DefinitionList
+
+
+class Definition_List_Item(DocutilsBodyNode):
     def build_flowable(self):
-        return rt.DefinitionList([item.process()
-                                  for item in self.definition_list_item])
-
-
-class Definition_List_Item(DocutilsBodySubNode):
-    def process(self):
-        term = self.term.styled_text()
+        term_text = self.term.styled_text()
         try:
-            term += ' : ' + self.classifier.styled_text()
+            for classifier in self.classifier:
+                term_text += ' : ' + classifier.styled_text()
         except AttributeError:
             pass
-        return (rt.DefinitionTerm([rt.Paragraph(term)]),
-                self.definition.flowable())
+        term = rt.StaticGroupedFlowables([rt.Paragraph(term_text)],
+                                         style='definition term')
+        return rt.LabeledFlowable(term, self.definition.flowable())
 
 
 class Term(DocutilsInlineNode):
@@ -469,12 +507,12 @@ class Classifier(DocutilsInlineNode):
 
 
 class Definition(DocutilsGroupingNode):
-    grouped_flowables_class = rt.Definition
+    style = 'definition'
 
 
-class Field_List(DocutilsBodyNode):
-    def build_flowable(self):
-        return rt.FieldList([field.flowable() for field in self.field])
+class Field_List(DocutilsGroupingNode):
+    grouped_flowables_class = rt.DefinitionList
+    style = 'field list'
 
 
 class Field(DocutilsBodyNode):
@@ -487,7 +525,7 @@ class Field(DocutilsBodyNode):
         return self.field_body.flowable()
 
     def build_flowable(self):
-        label = rt.Paragraph(self.field_name.styled_text(), style='field_name')
+        label = rt.Paragraph(self.field_name.styled_text(), style='field name')
         return rt.LabeledFlowable(label, self.field_body.flowable())
 
 
@@ -499,9 +537,9 @@ class Field_Body(DocutilsGroupingNode):
     pass
 
 
-class Option_List(DocutilsBodyNode):
-    def build_flowable(self):
-        return rt.FieldList([item.flowable() for item in self.option_list_item])
+class Option_List(DocutilsGroupingNode):
+    grouped_flowables_class = rt.DefinitionList
+    style = 'option list'
 
 
 class Option_List_Item(DocutilsBodyNode):
@@ -589,7 +627,7 @@ DOCUTILS_UNIT_TO_DIMENSION = {'': PT,    # assume points for unitless quantities
                               'mm': MM,
                               'pt': PT,
                               'pc': PICA,
-                              'px': DimensionUnit(1 / 100 * INCH),
+                              'px': DimensionUnit(1 / 100 * INCH, 'px'),
                               '%': PERCENT,
                               'em': None,
                               'ex': None}
@@ -605,7 +643,7 @@ def convert_quantity(quantity_string):
 class Table(DocutilsBodyNode):
     def build_flowable(self):
         tgroup = self.tgroup
-        if tgroup.get('colwidths', 'auto') == 'given':
+        if 'colwidths-given' in self.get('classes'):
             column_widths = [int(colspec.get('colwidth'))
                              for colspec in tgroup.colspec]
         else:
@@ -662,11 +700,29 @@ class Entry(DocutilsGroupingNode):
         return super().build_flowable(rowspan=rowspan, colspan=colspan)
 
 
-class Raw(DocutilsBodyNode):
+class Raw(DocutilsBodyNode, DocutilsInlineNode):
+    def build_styled_text(self):
+        cls, = self['classes']
+        return rt.WarnInline('{}: raw interpreted text roles are not supported'
+                             .format(cls))
+
     def build_flowable(self):
-        if self['format'] == 'pdf':   # rst2pdf
+        if self['format'] == 'rinoh':
+            # TODO: Flowable.from_text(self.text)
+            if self.text.startswith('ListOfFiguresSection'):
+                return rt.ListOfFiguresSection()
+            elif self.text == 'ListOfTablesSection':
+                return rt.ListOfTablesSection()
+            elif self.text == 'ListOfFigures(local=True)':
+                return rt.ListOfFigures(local=True)
+            elif self.text == 'ListOfTables(local=True)':
+                return rt.ListOfTables(local=True)
+            return rt.WarnFlowable("Unsupported raw pdf option: '{}'"
+                                   .format(self.text))
+        elif self['format'] == 'pdf':   # rst2pdf
             if self.text == 'PageBreak':
-                return rt.PageBreak(style=rt.PageBreakStyle(page_break=True))
+                style = rt.PageBreakStyle(page_break=rt.Break.ANY)
+                return rt.PageBreak(style=style)
             return rt.WarnFlowable("Unsupported raw pdf option: '{}'"
                                    .format(self.text))
         return rt.DummyFlowable()
