@@ -7,7 +7,7 @@
 
 """
 The layout engine. The container classes allow defining rectangular areas on a
-page to which :class:`Flowable`\ s can be rendered.
+page to which :class:`Flowable`\\ s can be rendered.
 
 * :class:`Container`: A rectangular area on a page to which flowables are
                       rendered.
@@ -39,7 +39,8 @@ __all__ = ['Container', 'FlowablesContainer', 'ChainedContainer',
            'DownExpandingContainer', 'InlineDownExpandingContainer',
            'UpExpandingContainer', 'VirtualContainer', 'Chain',
            'FootnoteContainer', 'MaybeContainer', 'discard_state',
-           'ContainerOverflow', 'EndOfContainer']
+           'ContainerOverflow', 'EndOfContainer',
+           'PageBreakException']
 
 
 class ContainerOverflow(Exception):
@@ -70,15 +71,16 @@ class ReflowRequired(Exception):
 
 
 class FlowableTarget(object):
-    """Something that takes :class:`Flowable`\ s to be rendered."""
+    """Something that takes :class:`Flowable`\\ s to be rendered."""
 
     def __init__(self, document_part, *args, **kwargs):
         """Initialize this flowable target.
 
         `document_part` is the :class:`Document` this flowable target is part
         of."""
-        self.flowables = []
-        document_part.flowable_targets.append(self)
+        from .flowable import StaticGroupedFlowables
+
+        self.flowables = StaticGroupedFlowables([])
         super().__init__(*args, **kwargs)
 
     @property
@@ -96,8 +98,7 @@ class FlowableTarget(object):
         return self
 
     def prepare(self, document):
-        for flowable in self.flowables:
-            flowable.prepare(self)
+        self.flowables.prepare(self)
 
 
 
@@ -185,8 +186,7 @@ class Container(object):
             child.render(type, rerender)
 
     def check_overflow(self):
-        for child in self.children:
-            child.check_overflow()
+        return all(child.check_overflow() for child in self.children)
 
     def place_children(self):
         for child in self.children:
@@ -206,10 +206,11 @@ class Container(object):
 BACKGROUND = 'background'
 CONTENT = 'content'
 HEADER_FOOTER = 'header_footer'
+CHAPTER_TITLE = 'chapter_title'
 
 
 class FlowablesContainerBase(Container):
-    """A :class:`Container` that renders :class:`Flowable`\ s to a rectangular
+    """A :class:`Container` that renders :class:`Flowable`\\ s to a rectangular
     area on a page. The first flowable is rendered at the top of the container.
     The next flowable is rendered below the first one, and so on."""
 
@@ -261,11 +262,25 @@ class FlowablesContainerBase(Container):
         elif ignore_overflow:
             self._self_cursor.grow(float(self.remaining_height))
         else:
-            raise ContainerOverflow
+            raise ContainerOverflow(self.page.number)
+
+    def advance2(self, height, ignore_overflow=False):
+        """Advance the cursor by `height`. Returns `True` on success.
+
+        Returns `False` if this would cause the cursor to point beyond the
+        bottom of the container.
+
+        """
+        if height <= self.remaining_height:
+            self._self_cursor.grow(height)
+        elif ignore_overflow:
+            self._self_cursor.grow(float(self.remaining_height))
+        else:
+            return False
+        return True
 
     def check_overflow(self):
-        if self.remaining_height < 0:
-            raise ReflowRequired
+        return self.remaining_height > 0
 
     def render(self, type, rerender=False):
         if type in (self.type, None):
@@ -296,9 +311,7 @@ class _FlowablesContainer(FlowableTarget, FlowablesContainerBase):
                          *args, **kwargs)
 
     def _render(self, type, rerender):
-        last_descender = None
-        for flowable in self.flowables:
-            _, _, last_descender = flowable.flow(self, last_descender)
+        self.flowables.flow(self, last_descender=None)
 
 
 class FlowablesContainer(_FlowablesContainer):
@@ -389,49 +402,24 @@ class ConditionalDownExpandingContainerBase(DownExpandingContainerBase):
             super().before_placing()
 
 
-class _InlineDownExpandingContainer(ConditionalDownExpandingContainerBase):
+class InlineDownExpandingContainer(ConditionalDownExpandingContainerBase):
+    """A :class:`DownExpandingContainer` whose top edge is placed at the
+    parent's current cursor position. As flowables are flowed in this container,
+    the parent's cursor also advances (but this behavior can be suppressed).
+
+    See :class:`Container` about the `name`, `parent`, `left`, `width`
+    and `right` parameters. Setting `advance_parent` to `False` prevents the
+    parent container's cursor being advanced.
+
+    """
+
     def __init__(self, name, parent, left=None, width=None, right=None,
-                 extra_space_below=0, advance_parent=True, place=True):
+                 advance_parent=True, place=True):
         super().__init__(name, None, parent, left=left, top=parent.cursor,
                          width=width, right=right,
                          max_height=parent.remaining_height, place=place)
         if advance_parent:
             parent._cursor.addends.append(self._cursor)
-        self.extra_space_below = extra_space_below
-
-    @property
-    def remaining_height(self):
-        return super().remaining_height - self.extra_space_below
-
-
-class InlineDownExpandingContainer(ContextManager):
-    """A :class:`DownExpandingContainer` whose top edge is placed at the
-    parent's current cursor position. As flowables are flowed in this container,
-    the part's cursor also advances (but this behavior can be suppressed).
-
-    This container can only be used as a context manager."""
-
-    def __init__(self, name, parent, left=None, width=None, right=None,
-                 extra_space_below=0, advance_parent=True):
-        """See :class:`Container` about on the `name`, `parent`, `left`, `width`
-        and `right` parameters.
-
-        `extra_space_below` specifies how much space should be reserved below
-        the flowables.
-
-        Setting `advance_parent` to `False` prevents the parent container's
-        cursor being advanced."""
-        self._container = _InlineDownExpandingContainer(name, parent,
-                                                        left, width, right,
-                                                        extra_space_below,
-                                                        advance_parent)
-
-    def __enter__(self):
-        return self._container
-
-    def __exit__(self, exc_type, exc_value, traceback):
-        if exc_value is None:
-            self._container.advance(self._container.extra_space_below, True)
 
 
 class UpExpandingContainer(_FlowablesContainer, ExpandingContainerBase):
@@ -452,7 +440,7 @@ class UpExpandingContainer(_FlowablesContainer, ExpandingContainerBase):
                          right=right, bottom=bottom, max_height=max_height)
 
 
-class _MaybeContainer(_InlineDownExpandingContainer):
+class _MaybeContainer(InlineDownExpandingContainer):
     def __init__(self, parent, left=None, width=None, right=None):
         super().__init__('MAYBE', parent, left=left, width=width, right=right,
                          place=False)
@@ -530,46 +518,37 @@ class FootnoteContainer(UpExpandingContainer):
         self.footnote_queue.append(footnote)
         if not self._flowing_footnotes:
             self._flowing_footnotes = True
-            try:
-                self.flow_footnotes()
-            finally:
-                self._flowing_footnotes = False
+            if not self.flow_footnotes():
+                return False
+            self._flowing_footnotes = False
+        return True
 
     def flow_footnotes(self):
         if self._reflowed:
             self._cursor.addends.pop()
             self._descenders.pop()
+        maybe_container = _MaybeContainer(self)
         while self.footnote_queue:
             footnote = self.footnote_queue.popleft()
             footnote_id = footnote.get_id(self.document)
             if footnote_id not in self.document.placed_footnotes:
-                with MaybeContainer(self) as maybe_container:
-                    _, _, descender = footnote.flow(maybe_container,
-                                                    self._descenders[-1])
-                    self._descenders.append(descender)
-                    self._reflowed = True
-                    self.page.check_overflow()
-                    self._reflowed = False
-                    self.document.placed_footnotes.add(footnote_id)
+                _, _, descender = footnote.flow(maybe_container,
+                                                self._descenders[-1])
+                self._descenders.append(descender)
+                self._reflowed = True
+                if not self.page.check_overflow():
+                    return False
+                self._reflowed = False
+                self.document.placed_footnotes.add(footnote_id)
+        maybe_container.do_place()
+        return True
+
     @property
     def next_number(self):
         self._footnote_number += 1
         return self._footnote_number
 
 
-class ChainState(object):
-    def __init__(self, flowable_index=0, flowable_state=None):
-        self.flowable_index = flowable_index
-        self.flowable_state = flowable_state
-
-    def __copy__(self):
-        return self.__class__(self.flowable_index, copy(self.flowable_state))
-
-    def next_flowable(self):
-        self.flowable_index += 1
-        self.flowable_state = None
-
-
 class Chain(FlowableTarget):
     """A :class:`FlowableTarget` that renders its flowables to a series of
     containers. Once a container is filled, the chain starts flowing flowables
@@ -583,13 +562,13 @@ class Chain(FlowableTarget):
         self.document_part = document_part
         self._init_state()
         self.containers = []
+        self.done = True
 
     def _init_state(self):
         """Reset the state of this chain: empty the list of containers, and zero
         the counter keeping track of which flowable needs to be rendered next.
         """
-        self._state = ChainState()
-        self._fresh_page_state = copy(self._state)
+        self._state = self._fresh_page_state = None
         self._rerendering = False
 
     @property
@@ -607,22 +586,18 @@ class Chain(FlowableTarget):
                 self._rerendering = True
         try:
             self.done = False
-            last_descender=None
-            while self._state.flowable_index < len(self.flowables):
-                flowable = self.flowables[self._state.flowable_index]
-                _, _, last_descender = flowable.flow(container, last_descender,
-                                                     self._state.flowable_state)
-                self._state.next_flowable()
+            self.flowables.flow(container, last_descender=None,
+                                state=self._state)
             # all flowables have been rendered
             if container == self.last_container:
                 self._init_state()    # reset state for the next rendering loop
             self.done = True
         except PageBreakException as exc:
-            self._state.flowable_state = exc.flowable_state
+            self._state = exc.flowable_state
             self._fresh_page_state = copy(self._state)
             raise
         except EndOfContainer as e:
-            self._state.flowable_state = e.flowable_state
+            self._state = e.flowable_state
             if container == self.last_container:
                 # save state for when ReflowRequired occurs
                 self._fresh_page_state = copy(self._state)
