@@ -17,53 +17,39 @@ Base classes and exceptions for styled document elements.
                                delegated to the parent :class:`Styled`
 """
 
-import re
+
+import string
 
 from ast import literal_eval
-from configparser import ConfigParser
 from collections import OrderedDict, namedtuple
 from itertools import chain
 from operator import attrgetter
+from pathlib import Path
 
-from .attribute import AttributesDictionary, RuleSet, Var
+from .attribute import (WithAttributes, AttributesDictionary,
+                        RuleSet, RuleSetFile, Configurable,
+                        DefaultValueException)
 from .element import DocumentElement
 from .resource import Resource
-from .util import cached, unique, all_subclasses, NotImplementedAttribute
+from .util import (cached, all_subclasses, NotImplementedAttribute,
+                   class_property)
 from .warnings import warn
 
 
 __all__ = ['Style', 'Styled', 'StyledMatcher', 'StyleSheet', 'StyleSheetFile',
-           'ClassSelector', 'ContextSelector', 'PARENT_STYLE', 'StyleException']
+           'ClassSelector', 'ContextSelector', 'PARENT_STYLE']
 
 
-class StyleException(Exception):
-    """Style lookup requires special handling."""
+class StyleMeta(WithAttributes):
+    def __new__(mcls, classname, bases, cls_dict):
+        if '__doc__' not in cls_dict:
+            styled_class_name = classname.replace('Style', '')
+            cls_dict['__doc__'] = ('Style class for :class:`.{}`'
+                                   .format(styled_class_name))
+        return super().__new__(mcls, classname, bases, cls_dict)
 
 
-class ParentStyleException(StyleException):
-    """Style attribute not found. Consult the parent :class:`Styled`."""
-
-
-class BaseStyleException(StyleException):
-    """The `attribute` is not specified in this :class:`Style`. Try to find the
-    attribute in a base style instead."""
-
-    def __init__(self, base_name, attribute):
-        self.base_name = base_name
-        self.attribute = attribute
-
-
-class DefaultStyleException(StyleException):
-    """The attribute is not specified in this :class:`Style` or any of its base
-    styles. Return the default value for the attribute."""
-
-
-class NoStyleException(StyleException):
-    """No style matching the given :class:`Styled` was found in the
-    :class:`StyleSheet`."""
-
-
-class Style(AttributesDictionary):
+class Style(AttributesDictionary, metaclass=StyleMeta):
     """Dictionary storing style attributes.
 
     The style attributes associated with this :class:`Style` are specified as
@@ -73,8 +59,6 @@ class Style(AttributesDictionary):
 
     """
 
-    default_base = None
-
     def __init__(self, base=None, **attributes):
         """Style attributes are as passed as keyword arguments. Supported
         attributes are the :class:`Attribute` class attributes of this style
@@ -87,6 +71,7 @@ class Style(AttributesDictionary):
         If `base` is a :class:`str`, it is used to look up the base style in
         the :class:`StyleSheet` this style is defined in."""
         super().__init__(base, **attributes)
+        self.name = None
 
     def __repr__(self):
         """Return a textual representation of this style."""
@@ -105,40 +90,17 @@ class Style(AttributesDictionary):
         else:
             return super().__getattr__(attribute)
 
-    def __getitem__(self, attribute):
-        """Return the value of `attribute`.
-
-        If the attribute is not specified in this :class:`Style`, find it in
-        this style's base styles (hierarchically), or ultimately raise a
-        :class:`DefaultValueException`."""
-        try:
-            return super().__getitem__(attribute)
-        except KeyError:
-            if self.base is None:
-                raise DefaultStyleException
-            elif isinstance(self.base, str):
-                raise BaseStyleException(self.base, attribute)
-            else:
-                return self.base[attribute]
+    @classmethod
+    def get_ruleset(self, document):
+        return document.stylesheet
 
 
 class ParentStyle(Style):
-    """Special style that delegates attribute lookups by raising a
-    :class:`ParentStyleException` on each attempt to access an attribute."""
-
-    def __repr__(self):
-        return self.__class__.__name__
-
-    def __getitem__(self, attribute):
-        raise ParentStyleException
-
-    def __bool__(self):
-        return True
+    """Style that forwards attribute lookups to the parent of the
+    :class:`Styled` from which the lookup originates."""
 
 
 PARENT_STYLE = ParentStyle()
-"""Special style that forwards style lookups to the parent of the
-:class:`Styled` from which the lookup originates."""
 
 
 class Selector(object):
@@ -165,10 +127,13 @@ class Selector(object):
     def __neg__(self):
         return self.pri(-1)
 
+    def __eq__(self, other):
+        return self.__dict__ == other.__dict__
+
     def pri(self, priority):
         return SelectorWithPriority(self, priority)
 
-    def get_styled_class(self, matcher):
+    def get_styled_class(self, stylesheet_or_matcher):
         raise NotImplementedError
 
     def get_style_name(self, matcher):
@@ -178,7 +143,10 @@ class Selector(object):
     def referenced_selectors(self):
         raise NotImplementedError
 
-    def match(self, styled, container):
+    def flatten(self, stylesheet):
+        raise NotImplementedError
+
+    def match(self, styled, stylesheet):
         raise NotImplementedError
 
 
@@ -190,8 +158,8 @@ class SelectorWithPriority(Selector):
     def pri(self, priority):
         return SelectorWithPriority(self.selector, self.priority + priority)
 
-    def get_styled_class(self, matcher):
-        return self.selector.get_styled_class(matcher)
+    def get_styled_class(self, stylesheet_or_matcher):
+        return self.selector.get_styled_class(stylesheet_or_matcher)
 
     def get_style_name(self, matcher):
         return self.selector.get_style_name(matcher)
@@ -204,12 +172,12 @@ class SelectorWithPriority(Selector):
     def referenced_selectors(self):
         return self.selector.referenced_selectors
 
-    def flatten(self, container):
-        flattened_selector = self.selector.flatten(container)
+    def flatten(self, stylesheet):
+        flattened_selector = self.selector.flatten(stylesheet)
         return flattened_selector.pri(self.priority)
 
-    def match(self, styled, container):
-        score = self.selector.match(styled, container)
+    def match(self, styled, stylesheet):
+        score = self.selector.match(styled, stylesheet)
         if score:
             score = Specificity(self.priority, 0, 0, 0, 0) + score
         return score
@@ -225,7 +193,7 @@ class EllipsisSelector(Selector):
         return
         yield
 
-    def flatten(self, container):
+    def flatten(self, stylesheet):
         return self
 
 
@@ -243,24 +211,24 @@ class SelectorByName(SingleSelector):
     def referenced_selectors(self):
         yield self.name
 
-    def flatten(self, container):
-        return container.document.stylesheet.get_selector(self.name)
+    def flatten(self, stylesheet):
+        return stylesheet.get_selector(self.name)
 
-    def get_styled_class(self, matcher):
-        selector = matcher.by_name[self.name]
-        return selector.get_styled_class(matcher)
+    def get_styled_class(self, stylesheet_or_matcher):
+        selector = stylesheet_or_matcher.get_selector(self.name)
+        return selector.get_styled_class(stylesheet_or_matcher)
 
     def get_style_name(self, matcher):
         selector = matcher.by_name[self.name]
         return selector.get_style_name(matcher)
 
-    def match(self, styled, container):
-        selector = container.document.stylesheet.get_selector(self.name)
-        return selector.match(styled, container)
+    def match(self, styled, stylesheet):
+        selector = stylesheet.get_selector(self.name)
+        return selector.match(styled, stylesheet)
 
 
 class ClassSelectorBase(SingleSelector):
-    def get_styled_class(self, matcher):
+    def get_styled_class(self, stylesheet_or_matcher):
         return self.cls
 
     @property
@@ -268,13 +236,13 @@ class ClassSelectorBase(SingleSelector):
         return
         yield
 
-    def flatten(self, container):
+    def flatten(self, stylesheet):
         return self
 
     def get_style_name(self, matcher):
         return self.style_name
 
-    def match(self, styled, container):
+    def match(self, styled, document):
         if not isinstance(styled, self.cls):
             return None
         class_match = 2 if type(styled) == self.cls else 1
@@ -313,18 +281,18 @@ class ContextSelector(Selector):
             for name in selector.referenced_selectors:
                 yield name
 
-    def flatten(self, container):
+    def flatten(self, stylesheet):
         return type(self)(*(child_selector for selector in self.selectors
                             for child_selector
-                            in selector.flatten(container).selectors))
+                            in selector.flatten(stylesheet).selectors))
 
-    def get_styled_class(self, matcher):
-        return self.selectors[-1].get_styled_class(matcher)
+    def get_styled_class(self, stylesheet_or_matcher):
+        return self.selectors[-1].get_styled_class(stylesheet_or_matcher)
 
     def get_style_name(self, matcher):
         return self.selectors[-1].get_style_name(matcher)
 
-    def match(self, styled, container):
+    def match(self, styled, stylesheet):
         def styled_and_parents(element):
             while element is not None:
                 yield element
@@ -339,20 +307,20 @@ class ContextSelector(Selector):
                 element = next(elements)                # NoMoreParentElement
                 if isinstance(selector, EllipsisSelector):
                     selector = next(selectors)          # StopIteration
-                    while not selector.match(element, container):
+                    while not selector.match(element, stylesheet):
                         element = next(elements)        # NoMoreParentElement
             except NoMoreParentElement:
                 return None
             except StopIteration:
                 break
-            score = selector.match(element, container)
+            score = selector.match(element, stylesheet)
             if not score:
                 return None
             total_score += score
         return total_score
 
 
-class NoMoreParentElement(StopIteration):
+class NoMoreParentElement(Exception):
     """The top-level element in the document element tree has been reached"""
 
 
@@ -380,10 +348,10 @@ class DocumentLocationSelector(object):
     def get_style_name(self, matcher):
         return self.selector.get_style_name(matcher)
 
-    def match(self, styled, container):
-        location_match = self.location_class.match(styled, container)
+    def match(self, styled, stylesheet):
+        location_match = self.location_class.match(styled, stylesheet)
         if location_match:
-            match = self.selector.match(styled, container)
+            match = self.selector.match(styled, stylesheet)
             if match:
                 return location_match + match
         return None
@@ -393,6 +361,9 @@ class StyledMeta(type, ClassSelectorBase):
     attributes = {}
     style_name = None
 
+    def __hash__(self):
+        return hash(id(self))
+
     @property
     def cls(cls):
         return cls
@@ -401,7 +372,7 @@ class StyledMeta(type, ClassSelectorBase):
         return ClassSelector(cls, style_name, **attributes)
 
 
-class Styled(DocumentElement, metaclass=StyledMeta):
+class Styled(DocumentElement, Configurable, metaclass=StyledMeta):
     """A document element who's style can be configured.
 
     Args:
@@ -411,6 +382,10 @@ class Styled(DocumentElement, metaclass=StyledMeta):
 
     """
 
+    @class_property
+    def configuration_class(cls):
+        return cls.style_class
+
     style_class = None
     """The :class:`Style` subclass that corresponds to this :class:`Styled`
     subclass."""
@@ -430,9 +405,10 @@ class Styled(DocumentElement, metaclass=StyledMeta):
                                     self.style_class.__name__,
                                     style.__class__.__name__))
         self.style = style
+        self.classes = []
 
     def __eq__(self, other):
-        return self.__dict__ == other.__dict__
+        return type(self) == type(other) and self.__dict__ == other.__dict__
 
     def __ne__(self, other):
         return not self == other
@@ -474,53 +450,54 @@ class Styled(DocumentElement, metaclass=StyledMeta):
         except AttributeError:
             return 0
 
-    @cached
-    def get_style(self, attribute, flowable_target):
+    def configuration_name(self, document):
         try:
-            return self.get_style_recursive(attribute, flowable_target)
-        except DefaultStyleException:
-            return self.style_class._get_default(attribute)
+            # raise AttributeError
+            return self._style_name
+        except AttributeError:
+            ruleset = self.configuration_class.get_ruleset(document)
+            self._style_name = ruleset.find_style(self)
+            return self._style_name
 
-    def get_base_style_recursive(self, exception, flowable_target):
-        stylesheet = flowable_target.document.stylesheet
-        try:
-            base_style = stylesheet[exception.base_name]
-            return base_style.get_value(exception.attribute, stylesheet)
-        except ParentStyleException:
-            return self.parent.get_style_recursive(exception.attribute,
-                                                   flowable_target)
-        except BaseStyleException as e:
-            return self.get_base_style_recursive(e, flowable_target)
-
-    def get_style_recursive(self, attribute, flowable_target):
-        stylesheet = flowable_target.document.stylesheet
-        try:
-            try:
-                style = self._style(flowable_target)
-                return style.get_value(attribute, stylesheet)
-            except NoStyleException:
-                if self.style_class.default_base == PARENT_STYLE:
-                    raise ParentStyleException
-                else:
-                    raise DefaultStyleException
-        except ParentStyleException:
-            parent = self.parent
-            try:
-                return parent.get_style(attribute, flowable_target)
-            except KeyError:  # 'attribute' is not supported by the parent
-                return parent.get_style_recursive(attribute, flowable_target)
-        except BaseStyleException as exception:
-            return self.get_base_style_recursive(exception, flowable_target)
+    def fallback_to_parent(self, attribute):
+        return isinstance(self.style, ParentStyle)
 
     @cached
-    def _style(self, container):
+    def get_style(self, attribute, container):
         if isinstance(self.style, Style):
-            return self.style
-        else:
-            return container.document.stylesheet.find_style(self, container)
+            try:
+                return self.style[attribute]
+            except KeyError:
+                pass
+        return self.get_config_value(attribute, container.document)
+
+    @property
+    def has_class(self):
+        return HasClass(self)
+
+    @property
+    def has_classes(self):
+        return HasClasses(self)
 
     def before_placing(self, container):
-        pass
+        if self.parent:
+            self.parent.before_placing(container)
+
+
+class HasClass(object):
+    def __init__(self, styled):
+        self.styled = styled
+
+    def __eq__(self, class_name):
+        return class_name in self.styled.classes
+
+
+class HasClasses(object):
+    def __init__(self, styled):
+        self.styled = styled
+
+    def __eq__(self, class_names):
+        return set(class_names).issubset(self.styled.classes)
 
 
 class InvalidStyledMatcher(Exception):
@@ -541,7 +518,7 @@ class StyledMatcher(dict):
 
     def __init__(self, mapping_or_iterable=None, **kwargs):
         super().__init__()
-        self.by_name = {}
+        self.by_name = OrderedDict()
         self._pending = {}
         self.update(mapping_or_iterable, **kwargs)
 
@@ -552,7 +529,7 @@ class StyledMatcher(dict):
     def __setitem__(self, name, selector):
         assert name not in self
         is_pending = False
-        for referenced_name in unique(selector.referenced_selectors):
+        for referenced_name in set(selector.referenced_selectors):
             if referenced_name not in self.by_name:
                 pending_selectors = self._pending.setdefault(referenced_name, {})
                 pending_selectors[name] = selector
@@ -568,6 +545,9 @@ class StyledMatcher(dict):
         if newly_defined_name in self._pending:
             self.update(self._pending.pop(newly_defined_name))
 
+    def get_selector(self, name):
+        return self.by_name[name]
+
     def check_validity(self):
         if self._pending:
             raise InvalidStyledMatcher(list(self._pending.keys()))
@@ -576,15 +556,15 @@ class StyledMatcher(dict):
         for name, selector in dict(iterable or (), **kwargs).items():
             self[name] = selector
 
-    def match(self, styled, container):
+    def match(self, styled, stylesheet):
         for cls in type(styled).__mro__:
             if cls not in self:
                 continue
             style_str = styled.style if isinstance(styled.style, str) else None
             for style in set((style_str, None)):
                 for name, selector in self[cls].get(style, {}).items():
-                    selector = selector.flatten(container)
-                    specificity = selector.match(styled, container)
+                    selector = selector.flatten(stylesheet)
+                    specificity = selector.match(styled, stylesheet)
                     if specificity:
                         yield Match(name, specificity)
 
@@ -592,49 +572,103 @@ class StyledMatcher(dict):
 class StyleSheet(RuleSet, Resource):
     """Dictionary storing a collection of related styles by name.
 
-    :class:`Style`\ s stored in a :class:`StyleSheet` can refer to their base
+    :class:`Style`\\ s stored in a :class:`StyleSheet` can refer to their base
     style by name.
 
     Args:
         name (str): a label for this style sheet
         matcher (StyledMatcher): the matcher providing the selectors the styles
-            contained in this style sheet map to. If no matcher is given, the
-            `base`\ 's matcher is used.
-        base (StyleSheet): the style sheet to extend
+            contained in this style sheet map to. If no matcher is given and
+            `base` is specified, the `base`\\ 's matcher is used. If `base` is
+            not set, the default matcher is used.
+        base (StyleSheet or str): the style sheet to extend
         description (str): a short string describing this style sheet
         pygments_style (str): the Pygments style to use for styling code blocks
 
     """
 
     resource_type = 'stylesheet'
+    main_section = 'STYLESHEET'
+    extension = '.rts'
 
     def __init__(self, name, matcher=None, base=None, description=None,
                  pygments_style=None, **user_options):
-        base = self.from_string(base) if isinstance(base, str) else base
         from .highlight import pygments_style_to_stylesheet
+        from .stylesheets import matcher as default_matcher
+
+        base = self.from_string(base) if isinstance(base, str) else base
+        if matcher is None:
+            matcher = default_matcher if base is None else StyledMatcher()
+        if matcher is not None:
+            matcher.check_validity()
         if pygments_style:
             base = pygments_style_to_stylesheet(pygments_style, base)
-        super().__init__(base)
-        self.name = name
+        super().__init__(name, base=base)
         self.description = description
-        self.matcher = matcher if matcher is not None else StyledMatcher()
-        self.matcher.check_validity()
+        self.matcher = matcher
         if user_options:
             warn('Unsupported options passed to stylesheet: {}'
                  .format(', '.join(user_options.keys())))
         self.user_options = user_options
-        self.variables = {}
+
+    def __str__(self):
+        for name, entry_point in self.installed_resources:
+            if self is entry_point.load():
+                return name
+        raise NotImplementedError
+
+    @classmethod
+    def parse_string(cls, filename_or_resource_name):
+        stylesheet_path = Path(filename_or_resource_name)
+        if stylesheet_path.is_file():
+            return StyleSheetFile(filename_or_resource_name)
+        else:
+            return super().parse_string(filename_or_resource_name)
+
+    @classmethod
+    def doc_repr(cls, value):
+        for name, entry_point in cls.installed_resources:
+            if value is entry_point.load():
+                object_name, = entry_point.attrs
+                return ('``{}`` (= :data:`{}.{}`)'
+                        .format(name, entry_point.module_name, object_name))
+        raise NotImplementedError
+
+    @classmethod
+    def doc_format(cls):
+        return ('the name of an :ref:`installed style sheet <included_style_'
+                'sheets>` or the filename of a stylesheet file (with the '
+                '``{}`` extension)'.format(cls.extension))
+
+    def _get_value_lookup(self, styled, attribute, document):
+        try:
+            return super()._get_value_lookup(styled, attribute, document)
+        except DefaultValueException:
+            if styled.fallback_to_parent(attribute):
+                return self._get_value_lookup(styled.parent, attribute, document)
+            raise
 
     def get_styled(self, name):
-        return self.get_selector(name).get_styled_class(self.matcher)
+        return self.get_selector(name).get_styled_class(self)
 
     def get_entry_class(self, name):
         return self.get_styled(name).style_class
 
-    def _get_variable(self, name, accepted_type):
-        return self.variables[name]
-
     def get_selector(self, name):
+        """Find a selector mapped to a style in this or a base style sheet.
+
+        Args:
+            name (str): a style name
+
+        Returns:
+            :class:`.Selector`: the selector mapped to the style `name`
+
+        Raises:
+            KeyError: if the style `name` was not found in this or a base
+                style sheet
+
+        """
+
         try:
             return self.matcher.by_name[name]
         except (AttributeError, KeyError):
@@ -643,28 +677,60 @@ class StyleSheet(RuleSet, Resource):
             else:
                 raise KeyError("No selector found for style '{}'".format(name))
 
-    def find_matches(self, styled, container):
-        for match in self.matcher.match(styled, container):
+    def find_matches(self, styled):
+        for match in self.matcher.match(styled, self):
             yield match
         if self.base is not None:
-            for match in self.base.find_matches(styled, container):
-                yield match
+            yield from self.base.find_matches(styled)
 
-    def find_style(self, styled, container):
-        matches = sorted(self.find_matches(styled, container),
+    def find_style(self, styled):
+        matches = sorted(self.find_matches(styled),
                          key=attrgetter('specificity'), reverse=True)
-        if len(set(match.specificity for match in matches)) < len(matches):
-            styled.warn('Multiple selectors match with the same specificity. '
-                        'See the style log for details.', container)
+        last_match = Match(None, ZERO_SPECIFICITY)
         for match in matches:
-            try:
-                return self[match.style_name]
-            except KeyError:
-                pass
-        raise NoStyleException
+            if (match.specificity == last_match.specificity
+                    and match.style_name != last_match.style_name):
+                styled.warn('Multiple selectors match with the same '
+                            'specificity. See the style log for details.')
+            if self.contains(match.style_name):
+                return match.style_name
+            last_match = match
+        raise DefaultValueException
+
+    def write(self, base_filename):
+        from configparser import ConfigParser
+        config = ConfigParser(interpolation=None)
+        config.add_section(self.main_section)
+        main = config[self.main_section]
+        main['name'] = self.name
+        main['description'] = self.description or ''
+
+        config.add_section('VARIABLES')
+        variables = config['VARIABLES']
+        for name, value in self.variables.items():
+            variables[name] = str(value)
+
+        for style_name, style in self.items():
+            classifier = ('' if style_name in self.matcher.by_name
+                          else ':' + type(style).__name__.replace('Style', ''))
+            config.add_section(style_name + classifier)
+            section = config[style_name + classifier]
+            section['base'] = str(style.base)
+            for name in type(style).supported_attributes:
+                try:
+                    section[name] = str(style[name])
+                except KeyError:  # default
+                    section[';' + name] = str(style._get_default(name))
+        with open(base_filename + self.extension, 'w') as file:
+            config.write(file, space_around_delimiters=True)
+            print(';Undefined styles:', file=file)
+            for style_name, selector in self.matcher.by_name.items():
+                if style_name in self:
+                    continue
+                print(';[{}]'.format(style_name), file=file)
 
 
-class StyleSheetFile(StyleSheet):
+class StyleSheetFile(RuleSetFile, StyleSheet):
     """Loads styles defined in a `.rts` file (INI format).
 
     Args:
@@ -678,167 +744,167 @@ class StyleSheetFile(StyleSheet):
 
     """
 
-    RE_VARIABLE = re.compile(r'^\$\(([a-z_ -]+)\)$', re.IGNORECASE)
-    RE_SELECTOR = re.compile(r'^(?P<name>[a-z]+)\((?P<args>.*)\)$', re.I)
-
-    def __init__(self, filename, matcher=None, base=None, **kwargs):
-        config = ConfigParser(default_section=None, delimiters=('=',),
-                              comment_prefixes=('#', ), interpolation=None)
-        with open(filename) as file:
-            config.read_file(file)
-        options = dict(config['STYLESHEET']
-                       if config.has_section('STYLESHEET') else {})
-        name = options.pop('name', filename)
-        base = options.pop('base', base)
-        options.update(kwargs)    # optionally override options
-        super().__init__(name, matcher, base, **options)
-        self.filename = filename
-        if config.has_section('VARIABLES'):
-            for name, value in config.items('VARIABLES'):
-                self.variables[name] = value
-        for section_name, section_body in config.items():
-            if section_name in (None, 'STYLESHEET', 'VARIABLES'):
-                continue
+    def process_section(self, style_name, selector, items):
+        if selector:
+            selector = parse_selector(selector)
+            styled_class = selector.get_styled_class(self)
+            if not isinstance(selector, StyledMeta):
+                self.matcher[style_name] = selector
             try:
-                style_name, selector  = section_name.split(':')
-                m = self.RE_SELECTOR.match(selector)
-                if m:
-                    styled_name = m.group('name')
-                    selector_args = m.group('args')
-                else:
-                    styled_name = selector
-                    selector_args = None
-                for styled_class in all_subclasses(Styled):
-                    if styled_class.__name__ == styled_name:
-                        style_cls = styled_class.style_class
-                        break
-                else:
-                    raise TypeError("Invalid type '{}' given for style '{}'"
-                                    .format(styled_name, style_name))
-                if selector_args:
-                    args, kwargs = parse_selector_args(selector_args)
-                    selector = styled_class.like(*args, **kwargs)
-                    self.matcher[style_name] = selector
-                try:
-                    matcher_styled = self.get_styled(style_name)
-                    if styled_class is not matcher_styled:
-                        raise TypeError("The type '{}' specified for style "
-                                        "'{}' does not match the type '{}' "
-                                        "returned by the matcher. Note that "
-                                        "you do not have to specify the type "
-                                        "in this case!"
-                                        .format(styled_class.__name__,
-                                                style_name,
-                                                matcher_styled.__name__))
-                except KeyError:
-                    pass
-            except ValueError:
-                style_name = section_name
-                style_cls = self.get_entry_class(style_name)
-            attribute_values = {}
-            for name, value in section_body.items():
-                value = value.replace('\n', ' ')
-                if name == 'base':
-                    attribute_values[name] = value
-                else:
-                    try:
-                        attribute = style_cls.attribute_definition(name)
-                    except KeyError:
-                        raise TypeError("'{}' is not a supported attribute for "
-                                        "'{}' ({})".format(name, style_name,
-                                                           style_cls.__name__))
-                    stripped = value.strip()
-                    m = self.RE_VARIABLE.match(stripped)
-                    if m:
-                        variable_name, = m.groups()
-                        value = Var(variable_name)
-                    else:
-                        accepted_type = attribute.accepted_type
-                        value = accepted_type.from_string(stripped)
-                    attribute_values[name] = value
-            self[style_name] = style_cls(**attribute_values)
-
-    def _get_variable(self, name, accepted_type):
-        return accepted_type.from_string(self.variables[name])
+                matcher_styled = self.get_styled(style_name)
+                if styled_class is not matcher_styled:
+                    raise TypeError("The type '{}' specified for style "
+                                    "'{}' does not match the type '{}' "
+                                    "returned by the matcher. Note that "
+                                    "you do not have to specify the type "
+                                    "in this case!"
+                                    .format(selector.__name__,
+                                            style_name,
+                                            matcher_styled.__name__))
+            except KeyError:
+                pass
+            style_cls = styled_class.style_class
+        else:
+            style_cls = self.get_entry_class(style_name)
+        self[style_name] = style_cls(**dict(items))
 
 
 class StyleParseError(Exception):
     pass
 
 
-def parse_selector_args(selector_args):
+def parse_selector(string):
+    chars = CharIterator(string)
+    selectors = []
+    while True:
+        eat_whitespace(chars)
+        first_char = chars.peek()
+        if first_char in ("'", '"'):
+            selector_name = parse_string(chars)
+            selector = SelectorByName(selector_name)
+        elif first_char == '.':
+            assert next(chars) + next(chars) + next(chars) == '...'
+            selector = EllipsisSelector()
+        else:
+            priority = 0
+            while chars.peek() in '+-':
+                priority += 1 if next(chars) == '+' else -1
+            selector = parse_class_selector(chars)
+            if priority != 0:
+                selector = selector.pri(priority)
+        selectors.append(selector)
+        eat_whitespace(chars)
+        try:
+            next(chars) == '/'
+        except StopIteration:
+            break
+    if len(selectors) == 1:
+        return selectors[0]
+    else:
+        return ContextSelector(*selectors)
+
+
+def parse_class_selector(chars):
+    styled_chars = []
+    eat_whitespace(chars)
+    while chars.peek() and chars.peek() in string.ascii_letters:
+        styled_chars.append(next(chars))
+    has_args = chars.peek() == '('
+    styled_name = ''.join(styled_chars)
+    for selector in all_subclasses(Styled):
+        if selector.__name__ == styled_name:
+            break
+    else:
+        raise TypeError("Invalid styled class '{}'".format(styled_name))
+    if has_args:
+        args, kwargs = parse_selector_args(chars)
+        selector = selector.like(*args, **kwargs)
+    return selector
+
+
+def parse_selector_args(chars):
     args, kwargs = [], {}
-    chars = CharIterator(selector_args)
-    try:
-        while True:
+    assert next(chars) == '('
+    eat_whitespace(chars)
+    while chars.peek() not in (None, ')'):
+        argument = parse_value(chars)
+        if argument is not None:
+            assert not kwargs
+            args.append(argument)
+        else:
+            keyword = parse_keyword(chars)
             eat_whitespace(chars)
-            argument = parse_value(chars)
-            if argument is not None:
-                assert not kwargs
-                args.append(argument)
-            else:
-                keyword = parse_keyword(chars)
-                eat_whitespace(chars)
-                kwargs[keyword] = parse_value(chars)
-            eat_whitespace(chars)
-            assert next(chars) == ','
-    except StopIteration:
-        pass
+            kwargs[keyword] = parse_value(chars)
+        eat_whitespace(chars)
+        char = next(chars)
+        if char == ')':
+            break
+        assert char == ','
+        eat_whitespace(chars)
+    else:
+        assert next(chars) == ')'
     return args, kwargs
 
 
 def eat_whitespace(chars):
-    for char in chars:
-        if char not in ' \t':
-            chars.push_back(char)
-            break
+    while chars.peek() and chars.peek() in ' \t':
+        next(chars)
 
 
-class CharIterator(object):
+class CharIterator(str):
     def __init__(self, string):
-        self._iter = iter(string)
-        self._pushed_back = []
+        self.next_index = 0
 
     def __iter__(self):
         return self
 
     def __next__(self):
-        if self._pushed_back:
-            return self._pushed_back.pop(0)
-        return next(self._iter)
+        index = self.next_index
+        self.next_index += 1
+        try:
+            return self[index]
+        except IndexError:
+            raise StopIteration
 
-    def push_back(self, char):
-        self._pushed_back.insert(0, char)
+    def match(self, chars):
+        """Return all next characters that are listed in `chars` as a string"""
+        start_index = self.next_index
+        for char in self:
+            if char not in chars:
+                self.next_index -= 1
+                break
+        return self[start_index:self.next_index]
+
+    def peek(self):
+        try:
+            return self[self.next_index]
+        except IndexError:
+            return None
 
 
 def parse_keyword(chars):
-    keyword_chars = []
-    for char in chars:
-        if not (char.isalnum() or char == '_'):
-            break
-        keyword_chars.append(char)
-    try:
-        while char != '=':
-            assert char in ' \t'
-            char = next(chars)
-    except (StopIteration, AssertionError):
+    keyword = chars.match(string.ascii_letters + string.digits + '_')
+    eat_whitespace(chars)
+    if chars.peek() != '=':
         raise StyleParseError('Expecting an equals sign to follow a keyword')
-    return ''.join(keyword_chars)
+    next(chars)
+    return keyword
 
 
 def parse_value(chars):
-    first_char = next(chars)
+    first_char = chars.peek()
     if first_char in ("'", '"'):
-        argument = parse_string(first_char, chars)
+        value = parse_string(chars)
     elif first_char.isnumeric() or first_char in '+-':
-        argument = parse_number(first_char, chars)
+        value = parse_number(chars)
     else:
-        chars.push_back(first_char)
-        argument = None
-    return argument
+        value = None
+    return value
 
 
-def parse_string(open_quote, chars):
+def parse_string(chars):
+    open_quote = next(chars)
+    assert open_quote in '"\''
     string_chars = [open_quote]
     escape_next = False
     for char in chars:
@@ -855,14 +921,8 @@ def parse_string(open_quote, chars):
     return literal_eval(''.join(string_chars))
 
 
-def parse_number(first_char, chars):
-    number_chars = [first_char]
-    for char in chars:
-        if char not in '0123456789.e+-':
-            chars.push_back(char)
-            break
-        number_chars.append(char)
-    return literal_eval(''.join(number_chars))
+def parse_number(chars):
+    return literal_eval(chars.match('0123456789.e+-'))
 
 
 class Specificity(namedtuple('Specificity',
@@ -912,7 +972,7 @@ class StyleLog(object):
         self.entries = []
 
     def log_styled(self, styled, container, continued, custom_message=None):
-        matches = self.stylesheet.find_matches(styled, container)
+        matches = self.stylesheet.find_matches(styled)
         log_entry = StyleLogEntry(styled, container, matches, continued,
                                   custom_message)
         self.entries.append(log_entry)
@@ -921,7 +981,8 @@ class StyleLog(object):
         raise NotImplementedError
 
     def write_log(self, filename_root):
-        with open(filename_root + '.stylelog', 'w') as log:
+        stylelog_path = filename_root.with_suffix('.stylelog')
+        with stylelog_path.open('w', encoding='utf-8') as log:
             current_page = None
             current_container = None
             for entry in self.entries:
@@ -964,15 +1025,15 @@ class StyleLog(object):
                                  key=attrgetter('specificity'))
                 if matches:
                     for match in matches:
-                        try:
-                            _ = self.stylesheet[match.style_name]
+                        if self.stylesheet.contains(match.style_name):
                             if first:
-                                style = _
+                                name = match.style_name
+                                style = self.stylesheet.get_configuration(name)
                                 label = '>'
                                 first = False
                             else:
                                 label = ' '
-                        except KeyError:
+                        else:
                             label = 'x'
                         specificity = ','.join(str(score)
                                                for score in match.specificity)
