@@ -6,7 +6,9 @@
 # Public License v3. See the LICENSE file or http://www.gnu.org/licenses/.
 
 
+import inspect
 import os
+import sys
 
 from .. import DATA_PATH
 from ..style import StyleSheetFile
@@ -19,12 +21,22 @@ __all__ = ['matcher', 'sphinx', 'sphinx_base14']
 
 STYLESHEETS_PATH = os.path.join(DATA_PATH, 'stylesheets')
 
+
 def path(filename):
     return os.path.join(STYLESHEETS_PATH, filename)
 
 
-sphinx = StyleSheetFile(path('sphinx.rts'), matcher)
+sphinx = StyleSheetFile(path('sphinx.rts'))
 
 sphinx_article = StyleSheetFile(path('sphinx_article.rts'))
 
 sphinx_base14 = StyleSheetFile(path('base14.rts'))
+
+
+# generate docstrings for the StyleSheet instances
+
+for name, stylesheet in inspect.getmembers(sys.modules[__name__]):
+    if not isinstance(stylesheet, StyleSheetFile):
+        continue
+    stylesheet.__doc__ = ('{}\n\nEntry point name: ``{}``'
+                          .format(stylesheet.description, stylesheet))
