@@ -6,39 +6,38 @@
 # Public License v3. See the LICENSE file or http://www.gnu.org/licenses/.
 
 
-from collections import Iterable
+from collections.abc import Iterable
 from itertools import chain
 from functools import partial
 from math import sqrt
 
 from .attribute import Attribute, OptionSet, OverrideDefault, Integer, Bool
 from .dimension import DimensionBase as DimBase
-from .draw import Line, Rectangle, ShapeStyle
-from .flowable import (HorizontallyAlignedFlowable,
-                       HorizontallyAlignedFlowableStyle,
-                       HorizontallyAlignedFlowableState)
+from .draw import Line, Rectangle, ShapeStyle, LineStyle
+from .flowable import Flowable, FlowableStyle, FlowableState, FlowableWidth
 from .layout import MaybeContainer, VirtualContainer, EndOfContainer
-from .reference import NUMBER
-from .structure import StaticGroupedFlowables, GroupedFlowablesStyle
+from .structure import (StaticGroupedFlowables, GroupedFlowablesStyle,
+                        ListOf, ListOfSection)
 from .style import Styled
-from .util import ReadAliasAttribute
+from .util import ReadAliasAttribute, NotImplementedAttribute
 
 
-__all__ = ['Table', 'TableWithCaption',
+__all__ = ['Table', 'TableStyle', 'TableWithCaption',
            'TableSection', 'TableHead', 'TableBody', 'TableRow',
-           'TableCell', 'TableCellStyle', 'TableCellBorder',
-           'TableCellBackground',
-           'TOP', 'MIDDLE', 'BOTTOM']
+           'TableCell', 'TableCellStyle',
+           'TableCellBorder', 'TableCellBorderStyle',
+           'TableCellBackground', 'TableCellBackgroundStyle', 'VerticalAlign',
+           'ListOfTables', 'ListOfTablesSection']
 
 
-class TableState(HorizontallyAlignedFlowableState):
-    def __init__(self, table, column_widths, body_row_index=0):
+class TableState(FlowableState):
+    table = ReadAliasAttribute('flowable')
+
+    def __init__(self, table, column_widths=None, body_row_index=0):
         super().__init__(table)
         self.column_widths = column_widths
         self.body_row_index = body_row_index
 
-    table = ReadAliasAttribute('flowable')
-
     @property
     def width(self):
         return sum(self.column_widths)
@@ -57,7 +56,7 @@ class TableState(HorizontallyAlignedFlowableState):
                               self.body_row_index)
 
 
-class TableStyle(HorizontallyAlignedFlowableStyle):
+class TableStyle(FlowableStyle):
     split_minimum_rows = Attribute(Integer, 0, 'The minimum number of rows to '
                                                'display when the table is '
                                                'split across pages')
@@ -68,7 +67,7 @@ class TableStyle(HorizontallyAlignedFlowableStyle):
 NEVER_SPLIT = float('+inf')
 
 
-class Table(HorizontallyAlignedFlowable):
+class Table(Flowable):
     style_class = TableStyle
 
     def __init__(self, body, head=None, width=None, column_widths=None,
@@ -76,11 +75,11 @@ class Table(HorizontallyAlignedFlowable):
         """
 
         Args:
-          width (:class:`DimensionBase` or `None`): the width of the table. If
-              `None`, the with of the table is automatically determined.
-          column_widths (list or `None`): a list of relative (:class:`Number`)
-              and absolute (:class:`DimensionBase`) column widths. A value of
-              `None` auto-sizes the column. Passing `None` instead of a list
+          width (DimensionBase or None): the width of the table. If ``None``,
+              the width of the table is automatically determined.
+          column_widths (list or None): a list of relative (int or float)
+              and absolute (:class:`.Dimension`) column widths. A value of
+              ``None`` auto-sizes the column. Passing ``None` instead of a list
               auto-sizes all columns.
 
         """
@@ -93,18 +92,25 @@ class Table(HorizontallyAlignedFlowable):
         self.column_widths = column_widths
 
     def initial_state(self, container):
-        return TableState(self, self._size_columns(container))
+        return TableState(self)
 
-    def render(self, container, last_descender, state, **kwargs):
+    def render(self, container, last_descender, state, space_below=0,
+               **kwargs):
         # TODO: allow data to override style (align)
-        get_style = partial(self.get_style, flowable_target=container)
+        if state.column_widths is None:
+            state.column_widths = self._size_columns(container)
+        get_style = partial(self.get_style, container=container)
         with MaybeContainer(container) as maybe_container:
             def render_rows(section, next_row_index=0):
                 rows = section[next_row_index:]
-                for rendered_rows in self._render_section(container, rows,
-                                                          state.column_widths):
+                rendered_spans = self._render_section(container, rows,
+                                                      state.column_widths)
+                for rendered_rows, is_last_span in rendered_spans:
                     sum_row_heights = sum(row.height for row in rendered_rows)
-                    if sum_row_heights > maybe_container.remaining_height:
+                    remaining_height = maybe_container.remaining_height
+                    if isinstance(section, TableBody) and is_last_span:
+                        remaining_height -= space_below
+                    if sum_row_heights > remaining_height:
                         break
                     self._place_rows_and_render_borders(maybe_container,
                                                         rendered_rows)
@@ -160,28 +166,25 @@ class Table(HorizontallyAlignedFlowable):
         try:
             fixed_width = width.to_points(container.width)
         except AttributeError:
-            fixed_width = width or None
+            fixed_width = width or FlowableWidth.AUTO
         min_column_widths = calculate_column_widths(0)
         max_column_widths = calculate_column_widths(float('+inf'))
 
         # calculate relative column widths for auto-sized columns
-        auto_relative_colwidths = [sqrt(minimum * maximum) for minimum, maximum
-                                   in zip(min_column_widths, max_column_widths)]
+        auto_rel_colwidths = [sqrt(minimum * maximum) for minimum, maximum
+                              in zip(min_column_widths, max_column_widths)]
         column_widths = self.column_widths or [None for _ in max_column_widths]
         try:          # TODO: max() instead of min()?
             relative_factor = min(auto_relative_width / width
-                                 for width, auto_relative_width
-                                 in zip(column_widths, auto_relative_colwidths)
-                                 if width and not isinstance(width, DimBase))
+                                  for width, auto_relative_width
+                                  in zip(column_widths, auto_rel_colwidths)
+                                  if width and not isinstance(width, DimBase))
         except ValueError:
             relative_factor = 1
         column_widths = [auto_relative_width * relative_factor
                          if width is None else width
                          for width, auto_relative_width
-                         in zip(column_widths, auto_relative_colwidths)]
-
-        # TODO: if table fits in width without wrapping cells, set with of
-        #       auto-sized columns to max_width
+                         in zip(column_widths, auto_rel_colwidths)]
 
         # set min = max for columns with a fixed width
         total_fixed_cols_width = 0
@@ -194,8 +197,12 @@ class Table(HorizontallyAlignedFlowable):
             else:
                 total_portions += column_width
 
+        # does the table fit within the available width without wrapping?
+        if sum(max_column_widths) <= container.width:
+            return max_column_widths
+
         # determine table width
-        if fixed_width:
+        if fixed_width != FlowableWidth.AUTO:
             table_width = fixed_width
         elif total_portions:
             max_factor = max(maximum / width for width, maximum,
@@ -203,20 +210,51 @@ class Table(HorizontallyAlignedFlowable):
                              if not isinstance(width, DimBase))
             no_wrap_rel_cols_width = total_portions * max_factor
             table_width = min(total_fixed_cols_width + no_wrap_rel_cols_width,
-                              container.width)
+                              float(container.width))
         else:
             table_width = total_fixed_cols_width
 
+        # TODO: warn when a column's fixed width < min width
+
         # determine absolute width of columns with relative widths
         if total_portions:
             total_relative_cols_width = table_width - total_fixed_cols_width
             portion_width = total_relative_cols_width / total_portions
-            return [width.to_points(container.width)
-                    if isinstance(width, DimBase) else width * portion_width
-                    for width in column_widths]
-        else:
-            return [width.to_points(container.width) for width in column_widths]
+            auto_widths = [width if isinstance(width, DimBase)
+                           else width * portion_width
+                           for width in column_widths]
+            extra = [0 if isinstance(width, DimBase)
+                     else auto_width - min_column_widths[index]
+                     for index, auto_width in enumerate(auto_widths)]
+            excess = [0 if isinstance(width, DimBase)
+                      else auto_width - max_column_widths[index]
+                      for index, auto_width in enumerate(auto_widths)]
+            excess_pos = sum(x for x in excess if x > 0)
+            extra_neg = sum(x for x in extra if x <= 0) + excess_pos
+            extra_pos = (sum(x for x, c in zip(extra, excess)
+                             if x > 0 and c <= 0))
+            if extra_pos + extra_neg < 0:
+                self.warn('Table contents are too wide to fit within the '
+                          'available width', container)
+                return [width.to_points(container.width)
+                        if isinstance(width, DimBase) else width
+                        for width in auto_widths]
 
+            def final_column_width(index, width):
+                if isinstance(width, DimBase):
+                    return width.to_points(container.width)
+                elif excess[index] > 0:
+                    return max_column_widths[index]
+                elif extra[index] <= 0:
+                    return min_column_widths[index]
+                else:
+                    return width + (extra_neg * extra[index] / extra_pos)
+
+            return [final_column_width(i, width)
+                    for i, width in enumerate(auto_widths)]
+        else:
+            return [width.to_points(container.width)
+                    for width in column_widths]
 
     @classmethod
     def _render_section(cls, container, rows, column_widths):
@@ -227,7 +265,8 @@ class Table(HorizontallyAlignedFlowable):
             rendered_row = cls._render_row(column_widths, container, row)
             rendered_rows.append(rendered_row)
             if rows_left_in_span == 0:
-                yield cls._vertically_size_cells(rendered_rows)
+                is_last_span = row == rows[-1]
+                yield cls._vertically_size_cells(rendered_rows), is_last_span
                 rendered_rows = []
         assert not rendered_rows
 
@@ -291,11 +330,11 @@ class Table(HorizontallyAlignedFlowable):
                 cell_container.place_at(container, x_cursor, y_pos)
                 vertical_align = rendered_cell.cell.get_style('vertical_align',
                                                               container)
-                if vertical_align == TOP:
+                if vertical_align == VerticalAlign.TOP:
                     vertical_offset = 0
-                elif vertical_align == MIDDLE:
+                elif vertical_align == VerticalAlign.MIDDLE:
                     vertical_offset = (cell_height - rendered_cell.height) / 2
-                elif vertical_align == BOTTOM:
+                elif vertical_align == VerticalAlign.BOTTOM:
                     vertical_offset = (cell_height - rendered_cell.height)
                 y_offset = float(y_cursor + vertical_offset)
                 rendered_cell.container.place_at(container, x_cursor, y_offset)
@@ -306,14 +345,6 @@ class Table(HorizontallyAlignedFlowable):
 class TableWithCaption(StaticGroupedFlowables):
     category = 'Table'
 
-    def prepare(self, flowable_target):
-        super().prepare(flowable_target)
-        document = flowable_target.document
-        number = document.counters.setdefault(self.category, 1)
-        document.counters[self.category] += 1
-        for id in self.get_ids(document):
-            document.set_reference(id, NUMBER, str(number))
-
 
 class TableSection(Styled, list):
     def __init__(self, rows, style=None, parent=None):
@@ -378,18 +409,14 @@ class TableRow(Styled, list):
         return spanned_columns
 
 
-TOP = 'top'
-MIDDLE = 'middle'
-BOTTOM = 'bottom'
-
-
 class VerticalAlign(OptionSet):
-    values = TOP, MIDDLE, BOTTOM
+    values = 'top', 'middle', 'bottom'
 
 
 class TableCellStyle(GroupedFlowablesStyle):
-    vertical_align = Attribute(VerticalAlign, MIDDLE, 'Vertical alignment of '
-                               'the cell contents within the available space')
+    vertical_align = Attribute(VerticalAlign, 'middle',
+                               'Vertical alignment of the cell contents '
+                               'within the available space')
 
 
 class TableCell(StaticGroupedFlowables):
@@ -509,7 +536,7 @@ class RenderedRow(list):
         super().append(rendered_cell)
 
 
-class TableCellBorderStyle(ShapeStyle):
+class TableCellBorderStyle(LineStyle):
     stroke = OverrideDefault(None)
 
 
@@ -537,3 +564,11 @@ class TableCellBackgroundStyle(ShapeStyle):
 
 class TableCellBackground(Rectangle):
     style_class = TableCellBackgroundStyle
+
+
+class ListOfTables(ListOf):
+    category = 'Table'
+
+
+class ListOfTablesSection(ListOfSection):
+    list_class = ListOfTables
