@@ -11,7 +11,3 @@ from .book import Book
 
 
 __all__ = ['Article', 'Book']
-
-
-ArticleConfiguration = Article.Configuration
-BookConfiguration = Book.Configuration
