@@ -8,22 +8,12 @@
 """
 Classes for describing styled text:
 
+* :class:`StyledText`: Base class for styled text.
 * :class:`SingleStyledText`: Text of a single style.
 * :class:`MixedStyledText`: Text where different substrings can have different
                             styles.
-* :class:`LiteralText`: Text that is typeset as is, including newlines and tabs.
 * :class:`TextStyle`: Style class specifying the font and other styling of text.
 
-A number of :class:`MixedStyledText` subclasses are provided for changing a
-single style attribute of the passed text:
-
-* :class:`Bold`
-* :class:`Italic`
-* :class:`Emphasized`
-* :class:`SmallCaps`
-* :class:`Superscript`
-* :class:`Subscript`
-
 Some characters with special properties and are represented by special classes:
 
 * :class:`Space`
@@ -39,24 +29,24 @@ import re
 
 from ast import literal_eval
 from html.entities import name2codepoint
-from itertools import groupby, tee
+from token import NAME, STRING, NEWLINE, LPAR, RPAR, ENDMARKER
 
 from .attribute import (AttributeType, Attribute, Bool, Integer,
-                        AcceptNoneAttributeType)
+                        AcceptNoneAttributeType, ParseError)
 from .color import Color, BLACK
-from .dimension import DimensionBase, PT
+from .dimension import Dimension, PT
 from .font import Typeface
 from .fonts import adobe14
 from .font.style import (FontWeight, FontSlant, FontWidth, FontVariant,
-                         TextPosition, MEDIUM, UPRIGHT, NORMAL, BOLD, ITALIC,
-                         SMALL_CAPITAL ,SUPERSCRIPT, SUBSCRIPT)
-from .style import Style, Styled, PARENT_STYLE, StyleException
+                         TextPosition)
+from .style import Style, Styled, StyledMeta
+from .util import NotImplementedAttribute
 
 
-__all__ = ['TextStyle', 'StyledText', 'SingleStyledText', 'MixedStyledText',
-           'ConditionalMixedStyledText', 'Space', 'FixedWidthSpace',
-           'NoBreakSpace', 'Spacer', 'Tab', 'Newline', 'Bold', 'Italic',
-           'Emphasized', 'SmallCaps', 'Superscript', 'Subscript']
+__all__ = ['TextStyle', 'StyledText', 'WarnInline', 'SingleStyledText',
+           'MixedStyledText', 'ConditionalMixedStyledText', 'Space',
+           'FixedWidthSpace', 'NoBreakSpace', 'Spacer', 'Tab', 'Newline',
+           'Superscript', 'Subscript']
 
 
 class Locale(AttributeType):
@@ -67,26 +57,29 @@ class Locale(AttributeType):
         return cls.REGEX.match(value) is not None
 
     @classmethod
-    def parse_string(cls, string):
-        if not cls.check_type(string):
+    def from_tokens(cls, tokens):
+        token = next(tokens)
+        if not cls.check_type(token.string):
             raise ValueError("'{}' is not a valid locale. Needs to be of the "
-                             "form 'en_US'.".format(string))
-        return string
+                             "form 'en_US'.".format(token.string))
+        return token.string
+
+    @classmethod
+    def doc_format(cls):
+        return 'locale identifier in the ``<language ID>_<region ID>`` format'
 
 
 class TextStyle(Style):
-    """The :class:`Style` for :class:`StyledText` objects"""
-
     typeface = Attribute(Typeface, adobe14.times, 'Typeface to set the text in')
-    font_weight = Attribute(FontWeight, MEDIUM, 'Thickness of character '
-                                                'outlines relative to their '
-                                                'height')
-    font_slant = Attribute(FontSlant, UPRIGHT, 'Slope style of the font')
-    font_width = Attribute(FontWidth, NORMAL, 'Stretch of the characters')
-    font_size = Attribute(DimensionBase, 10*PT, 'Height of characters')
+    font_weight = Attribute(FontWeight, 'medium', 'Thickness of character '
+                                                  'outlines relative to their '
+                                                  'height')
+    font_slant = Attribute(FontSlant, 'upright', 'Slope style of the font')
+    font_width = Attribute(FontWidth, 'normal', 'Stretch of the characters')
+    font_size = Attribute(Dimension, 10*PT, 'Height of characters')
     font_color = Attribute(Color, BLACK, 'Color of the font')
-    font_variant = Attribute(FontVariant, NORMAL, 'Variant of the font')
-    position = Attribute(TextPosition, NORMAL, 'Vertical text position')
+    font_variant = Attribute(FontVariant, 'normal', 'Variant of the font')
+    position = Attribute(TextPosition, 'normal', 'Vertical text position')
     kerning = Attribute(Bool, True, 'Improve inter-letter spacing')
     ligatures = Attribute(Bool, True, 'Run letters together where possible')
     # TODO: character spacing
@@ -97,8 +90,6 @@ class TextStyle(Style):
                                              'Accepts locale codes such as '
                                              "'en_US'")
 
-    default_base = PARENT_STYLE
-
 
 class CharacterLike(Styled):
     def __repr__(self):
@@ -115,7 +106,8 @@ class CharacterLike(Styled):
         raise NotImplementedError
 
 
-NAME2CHAR = {name: chr(codepoint) for name, codepoint in name2codepoint.items()}
+NAME2CHAR = {name: chr(codepoint)
+             for name, codepoint in name2codepoint.items()}
 
 
 class StyledText(Styled, AcceptNoneAttributeType):
@@ -143,55 +135,75 @@ class StyledText(Styled, AcceptNoneAttributeType):
         return super().check_type(value) or isinstance(value, str)
 
     @classmethod
-    def parse_string(cls, string):
-        def parse_text(chars):
-            text = quote = skip_whitespace(chars)
-            if not quote in ("'", '"'):
-                raise StyledTextParseError
-            escape = False
-            for char in chars:
-                text += char
-                if not escape and char == quote:
-                    return text
-                escape = char == '\\'
-            raise StyledTextParseError('Missing closing quote')
-
-        def parse_style(saved_chars):
-            chars, saved_chars = tee(saved_chars)
-            try:
-                next_char = skip_whitespace(chars)
-            except StopIteration:
-                return saved_chars, None
-            if next_char != '(':
-                return saved_chars, None
-            style = ''
-            for char in chars:
-                if char == ')':
-                    return chars, style
-                else:
-                    style += char
-            raise StyledTextParseError('Missing closing brace')
-
-        def skip_whitespace(chars):
-            for char in chars:
-                if char not in ' \t':
-                    return char
-            raise StopIteration
-
+    def from_tokens(cls, tokens):
         texts = []
-        chars = iter(string.strip())
-        while True:
-            try:
-                text = literal_eval(parse_text(chars))
-                chars, style = parse_style(chars)
+        for token in tokens:
+            if token.type == NEWLINE:
+                continue
+            elif token.type == NAME:      # inline flowable
+                directive = token.string.lower()
+                inline_flowable_class = InlineFlowable.directives[directive]
+                if next(tokens).exact_type != LPAR:
+                    raise ParseError('Expecting an opening parenthesis.')
+                args, kwargs = inline_flowable_class.arguments.parse_arguments(tokens)
+                if next(tokens).exact_type != RPAR:
+                    raise ParseError('Expecting a closing parenthesis.')
+                inline_flowable = inline_flowable_class(*args, **kwargs)
+                texts.append(inline_flowable)
+            elif token.type == STRING:    # text
+                text = literal_eval(token.string)
+                if tokens.next.exact_type == LPAR:
+                    _, start_col = next(tokens).end
+                    for token in tokens:
+                        if token.exact_type == RPAR:
+                            _, end_col = token.start
+                            style = token.line[start_col:end_col].strip()
+                            break
+                        elif token.type == NEWLINE:
+                            raise StyledTextParseError('No newline allowed in'
+                                                       'style name')
+                else:
+                    style = None
                 texts.append(cls._substitute_variables(text, style))
-            except StopIteration:
+            else:
+                raise StyledTextParseError('Expecting text or inline flowable')
+            if tokens.next.type == ENDMARKER:
                 break
-        return MixedStyledText(texts)
+
+        if len(texts) > 1:
+            return MixedStyledText(texts)
+        else:
+            first, = texts
+            return first
+
+    @classmethod
+    def doc_repr(cls, value):
+        return '``{}``'.format(value) if value else '(no value)'
+
+    @classmethod
+    def doc_format(cls):
+        return ("a list of styled text strings, separated by spaces. A styled "
+                "text string is a quoted string (``'`` or ``\"``), optionally "
+                "followed by a style name enclosed in braces: "
+                "``'text string' (style name)``")
+
+    @classmethod
+    def validate(cls, value, accept_variables=False, attribute_name=None):
+        if attribute_name is None and isinstance(value, str):
+            value = SingleStyledText(value)
+        return super().validate(value, accept_variables, attribute_name)
 
     @classmethod
     def _substitute_variables(cls, text, style):
-        return SingleStyledText(text.format(**NAME2CHAR), style=style)
+        def substitute_controlchars_htmlentities(string, style=None):
+            try:
+                return ControlCharacter.all[string]()
+            except KeyError:
+                return SingleStyledText(string.format(**NAME2CHAR),
+                                        style=style)
+
+        return Field.substitute(text, substitute_controlchars_htmlentities,
+                                style=style)
 
     def __str__(self):
         return self.to_string(None)
@@ -207,17 +219,17 @@ class StyledText(Styled, AcceptNoneAttributeType):
     def paragraph(self):
         return self.parent.paragraph
 
-    position = {SUPERSCRIPT: 1 / 3,
-                SUBSCRIPT: - 1 / 6}
+    def fallback_to_parent(self, attribute):
+        return attribute != 'position'
+
+    position = {TextPosition.SUPERSCRIPT: 1 / 3,
+                TextPosition.SUBSCRIPT: - 1 / 6}
     position_size = 583 / 1000
 
     def is_script(self, container):
         """Returns `True` if this styled text is super/subscript."""
-        try:
-            style = self._style(container)
-            return style.get_value('position', container) != NORMAL
-        except StyleException:
-            return False
+        position = self.get_style('position', container)
+        return position != TextPosition.NORMAL
 
     def script_level(self, container):
         """Nesting level of super/subscript."""
@@ -237,17 +249,18 @@ class StyledText(Styled, AcceptNoneAttributeType):
 
     def y_offset(self, container):
         """Vertical baseline offset (up is positive)."""
-        offset = (self.parent.y_offset(container)\
+        offset = (self.parent.y_offset(container)
                   if hasattr(self.parent, 'y_offset') else 0)
         if self.is_script(container):
-            style = self._style(container)
-            offset += (self.parent.height(container) *
-                       self.position[style.position])
-            # The Y offset should only change once for the nesting level
-            # where the position style is set, hence we don't recursively
-            # get the position style using self.get_style('position')
+            position = self.get_style('position', container)
+            offset += self.parent.height(container) * self.position[position]
         return offset
 
+    @property
+    def items(self):
+        """The list of items in this StyledText."""
+        raise NotImplementedError
+
     def spans(self, container):
         """Generator yielding all spans in this styled text, one
         item at a time (used in typesetting)."""
@@ -258,6 +271,30 @@ class StyledTextParseError(Exception):
     pass
 
 
+class WarnInline(StyledText):
+    """Inline element that emits a warning
+
+    Args:
+        message (str): the warning message to emit
+
+    """
+
+    def __init__(self, message, parent=None):
+        super().__init__(parent=parent)
+        self.message = message
+
+    def to_string(self, flowable_target):
+        return ''
+
+    @property
+    def items(self):
+        return [self]
+
+    def spans(self, container):
+        self.warn(self.message, container)
+        return iter([])
+
+
 class SingleStyledTextBase(StyledText):
     """Styled text where all text shares a single :class:`TextStyle`."""
 
@@ -298,40 +335,30 @@ class SingleStyledTextBase(StyledText):
         return (self.font(container).line_gap_in_pt
                 * float(self.get_style('font_size', container)))
 
+    @property
+    def items(self):
+        return [self]
+
     def spans(self, container):
         yield self
 
-    @staticmethod
-    def split_words(text):
-        def is_special_character(char):
-            return char in ' \t\n\N{ZERO WIDTH SPACE}'
 
-        for is_special, characters in groupby(text, is_special_character):
-            if is_special:
-                for char in characters:
-                    yield char
-            else:
-                yield ''.join(characters)
-
-    def split(self, container, **kwargs):
-        """Yield the words and spaces in this single-styled text."""
-        return self.split_words(self.text(container, **kwargs))
-
-    def before_placing(self, container):
-        pass
+ESCAPE = str.maketrans({"'": r"\'",
+                        '\n': r'\n',
+                        '\t': r'\t'})
 
 
 class SingleStyledText(SingleStyledTextBase):
     def __init__(self, text, style=None, parent=None):
-        """Initialize this single-styled text with `text` (:class:`str`),
-        `style`, and `parent` (see :class:`StyledText`).
-
-        In `text`, tab, line-feed and newline characters are all considered
-        whitespace. Consecutive whitespace characters are reduced to a single
-        space."""
         super().__init__(style=style, parent=parent)
         self._text = text
 
+    def __str__(self):
+        result = "'{}'".format(self._text.translate(ESCAPE))
+        if self.style is not None:
+            result += ' ({})'.format(self.style)
+        return result
+
     def text(self, container, **kwargs):
         return self._text
 
@@ -339,23 +366,33 @@ class SingleStyledText(SingleStyledTextBase):
 class MixedStyledTextBase(StyledText):
     def to_string(self, flowable_target):
         return ''.join(item.to_string(flowable_target)
-                       for item in self.text(flowable_target))
+                       for item in self.children(flowable_target))
 
     def spans(self, container):
         """Recursively yield all the :class:`SingleStyledText` items in this
         mixed-styled text."""
-        for item in self.text(container):
-            container.register_styled(item)
-            for span in item.spans(container):
+        for child in self.children(container):
+            container.register_styled(child)
+            for span in child.spans(container):
                 yield span
 
+    def children(self, flowable_target):
+        raise NotImplementedError
+
 
 class MixedStyledText(MixedStyledTextBase, list):
-    """Concatenation of :class:`StyledText` objects."""
+    """Concatenation of styled text
+
+    Args:
+        text_or_items (str, StyledText or iterable of these): mixed styled text
+        style: see :class:`.Styled`
+        parent: see :class:`.DocumentElement`
+
+    """
 
     _assumed_equal = {}
 
-    def __init__(self, text_or_items, style=None, parent=None):
+    def __init__(self, text_or_items, id=None, style=None, parent=None):
         """Initialize this mixed-styled text as the concatenation of
         `text_or_items`, which is either a single text item or an iterable of
         text items. Individual text items can be :class:`StyledText` or
@@ -363,19 +400,17 @@ class MixedStyledText(MixedStyledTextBase, list):
         each of the text items.
 
         See :class:`StyledText` for information on `style`, and `parent`."""
-        super().__init__(style=style, parent=parent)
+        super().__init__(id=id, style=style, parent=parent)
         if isinstance(text_or_items, (str, StyledText)):
             text_or_items = (text_or_items, )
         for item in text_or_items:
             self.append(item)
 
     def __add__(self, other):
-        if isinstance(other, str) or self.style == other.style:
-            try:
-                other_list = list(other)
-            except TypeError:
-                other_list = [other]
-            return MixedStyledText(list(self) + other_list,
+        if isinstance(other, str):
+            other = SingleStyledText(other)
+        if self.style == other.style:
+            return MixedStyledText(self.items + other.items,
                                    style=self.style, parent=self.parent)
         else:
             return super().__add__(other)
@@ -386,6 +421,10 @@ class MixedStyledText(MixedStyledTextBase, list):
         return '{}{} (style={})'.format(self.__class__.__name__,
                                         super().__repr__(), self.style)
 
+    def __str__(self):
+        assert self.style is None
+        return ' '.join(str(item) for item in self)
+
     def __eq__(self, other):
         # avoid infinite recursion due to the 'parent' attribute
         assumed_equal = type(self)._assumed_equal.setdefault(id(self), set())
@@ -412,8 +451,12 @@ class MixedStyledText(MixedStyledTextBase, list):
         item.parent = self
         list.append(self, item)
 
-    def text(self, flowable_target, **kwargs):
-        return self
+    @property
+    def items(self):
+        return list(self)
+
+    def children(self, flowable_target):
+        return self.items
 
 
 class ConditionalMixedStyledText(MixedStyledText):
@@ -499,14 +542,26 @@ class Box(Character):
         return self.width
 
 
-class ControlCharacter(Character):
+class ControlCharacterMeta(StyledMeta):
+    def __new__(metacls, classname, bases, cls_dict):
+        cls = super().__new__(metacls, classname, bases, cls_dict)
+        try:
+            ControlCharacter.all[cls.character] = cls
+        except NameError:
+            pass
+        return cls
+
+
+class ControlCharacter(Character, metaclass=ControlCharacterMeta):
     """A non-printing character that affects typesetting of the text near it."""
 
+    character = NotImplementedAttribute()
     exception = Exception
+    all = {}
 
-    def __init__(self, char):
+    def __init__(self):
         """Initialize this control character with it's unicode `char`."""
-        super().__init__(char)
+        super().__init__(self.character)
 
     def __repr__(self):
         """A textual representation of this control character."""
@@ -527,66 +582,23 @@ class NewlineException(Exception):
 class Newline(ControlCharacter):
     """Control character ending the current line and starting a new one."""
 
+    character = '\n'
     exception = NewlineException
 
-    def __init__(self, *args, **kwargs):
-        """Initiatize this newline character."""
-        super().__init__('\n')
-
 
 class Tab(ControlCharacter):
-    """Tabulator character, used for vertically aligning text."""
+    """Tabulator character, used for vertically aligning text.
 
-    def __init__(self, *args, **kwargs):
-        """Initialize this tab character. Its attribute :attr:`tab_width` is set
-        a later point in time when context (:class:`TabStop`) is available."""
-        super().__init__('\t')
+    The :attr:`tab_width` attribute is set a later point in time when context
+    (:class:`TabStop`) is available."""
+
+    character = '\t'
 
 
 # predefined text styles
 
-ITALIC_STYLE = EMPHASIZED_STYLE = TextStyle(font_slant=ITALIC)
-BOLD_STYLE = TextStyle(font_weight=BOLD)
-BOLD_ITALIC_STYLE = TextStyle(font_weight=BOLD, font_slant=ITALIC)
-SMALL_CAPITALS_STYLE = TextStyle(font_variant=SMALL_CAPITAL)
-SUPERSCRIPT_STYLE = TextStyle(position=SUPERSCRIPT)
-SUBSCRIPT_STYLE = TextStyle(position=SUBSCRIPT)
-
-
-class Bold(MixedStyledText):
-    """Bold text."""
-
-    def __init__(self, text):
-        """Accepts a single instance of :class:`str` or :class:`StyledText`, or
-        an iterable of these."""
-        super().__init__(text, style=BOLD_STYLE)
-
-
-class Italic(MixedStyledText):
-    """Italic text."""
-
-    def __init__(self, text, parent=None):
-        """Accepts a single instance of :class:`str` or :class:`StyledText`, or
-        an iterable of these."""
-        super().__init__(text, style=ITALIC_STYLE, parent=parent)
-
-
-class Emphasized(MixedStyledText):
-    """Emphasized text."""
-
-    def __init__(self, text, parent=None):
-        """Accepts a single instance of :class:`str` or :class:`StyledText`, or
-        an iterable of these."""
-        super().__init__(text, style=EMPHASIZED_STYLE, parent=parent)
-
-
-class SmallCaps(MixedStyledText):
-    """Small capitals text."""
-
-    def __init__(self, text, parent=None):
-        """Accepts a single instance of :class:`str` or :class:`StyledText`, or
-        an iterable of these."""
-        super().__init__(text, style=SMALL_CAPITALS_STYLE, parent=parent)
+SUPERSCRIPT_STYLE = TextStyle(position='superscript')
+SUBSCRIPT_STYLE = TextStyle(position='subscript')
 
 
 class Superscript(MixedStyledText):
@@ -605,3 +617,7 @@ class Subscript(MixedStyledText):
         """Accepts a single instance of :class:`str` or :class:`StyledText`, or
         an iterable of these."""
         super().__init__(text, style=SUBSCRIPT_STYLE, parent=parent)
+
+
+from .reference import Field
+from .inline import InlineFlowable
